"""Classes for report generation and add-ons."""

import logging
import os
import warnings

from copy import copy
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from jinja2 import Environment
from jinja2 import FileSystemLoader
from json2html import json2html
from sklearn.metrics import average_precision_score
from sklearn.metrics import confusion_matrix
from sklearn.metrics import explained_variance_score
from sklearn.metrics import f1_score
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import median_absolute_error
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import precision_score
from sklearn.metrics import r2_score
from sklearn.metrics import recall_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import roc_curve
from sklearn.utils.multiclass import type_of_target

from lightautoml.addons.uplift import metrics as uplift_metrics
from lightautoml.addons.uplift.metalearners import TLearner
from lightautoml.addons.uplift.metalearners import XLearner
from lightautoml.addons.uplift.utils import _get_treatment_role
from lightautoml.dataset import roles as laml_roles


logger = logging.getLogger(__name__)

base_dir = os.path.dirname(__file__)


def extract_params(input_struct):
    params = dict()
    iterator = input_struct if isinstance(input_struct, dict) else input_struct.__dict__
    for key in iterator:
        if key.startswith(("_", "autonlp_params")):
            continue
        value = iterator[key]
        if type(value) in [bool, int, float, str, np.bool_]:
            params[key] = value
        elif value is None:
            params[key] = None
        elif hasattr(value, "__dict__") or isinstance(value, dict):
            params[key] = extract_params(value)
        elif isinstance(value, list) or isinstance(value, np.ndarray):
            params[key] = ", ".join([i.__str__() for i in value])
        else:
            params[key] = str(type(value))
    return params


def plot_roc_curve_image(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    plt.figure(figsize=(10, 10))

    fpr, tpr, _ = roc_curve(data["y_true"], data["y_pred"])
    auc_score = roc_auc_score(data["y_true"], data["y_pred"])

    lw = 2
    plt.plot(fpr, tpr, color="blue", lw=lw, label="Trained model")
    plt.plot([0, 1], [0, 1], color="red", lw=lw, linestyle="--", label="Random model")
    plt.xlim([-0.05, 1.05])
    plt.ylim([-0.05, 1.05])
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    lgd = plt.legend(bbox_to_anchor=(0.5, -0.15), loc="upper center", ncol=2)
    plt.xticks(np.arange(0, 1.01, 0.05), rotation=45)
    plt.yticks(np.arange(0, 1.01, 0.05))
    plt.grid(color="gray", linestyle="-", linewidth=1)
    plt.title("ROC curve (GINI = {:.3f})".format(2 * auc_score - 1))
    plt.savefig(path, bbox_extra_artists=(lgd,), bbox_inches="tight")
    plt.close()
    return auc_score


def plot_pr_curve_image(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    plt.figure(figsize=(10, 10))

    precision, recall, _ = precision_recall_curve(data["y_true"], data["y_pred"])
    ap_score = average_precision_score(data["y_true"], data["y_pred"])

    lw = 2
    plt.plot(recall, precision, color="blue", lw=lw, label="Trained model")
    positive_rate = np.sum(data["y_true"] == 1) / data.shape[0]
    plt.plot(
        [0, 1],
        [positive_rate, positive_rate],
        color="red",
        lw=lw,
        linestyle="--",
        label="Random model",
    )
    plt.xlim([-0.05, 1.05])
    plt.ylim([0.45, 1.05])
    plt.xlabel("Recall")
    plt.ylabel("Precision")
    lgd = plt.legend(bbox_to_anchor=(0.5, -0.15), loc="upper center", ncol=2)
    plt.xticks(np.arange(0, 1.01, 0.05), rotation=45)
    plt.yticks(np.arange(0, 1.01, 0.05))
    plt.grid(color="gray", linestyle="-", linewidth=1)
    plt.title("PR curve (AP = {:.3f})".format(ap_score))
    plt.savefig(path, bbox_extra_artists=(lgd,), bbox_inches="tight")
    plt.close()


def plot_preds_distribution_by_bins(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, axs = plt.subplots(figsize=(16, 10))

    box_plot_data = []
    labels = []
    for name, group in data.groupby("bin"):
        labels.append(name)
        box_plot_data.append(group["y_pred"].values)

    box = axs.boxplot(box_plot_data, patch_artist=True, labels=labels)
    for patch in box["boxes"]:
        patch.set_facecolor("green")
    axs.set_yscale("log")
    axs.set_xlabel("Bin number")
    axs.set_ylabel("Prediction")
    axs.set_title("Distribution of object predictions by bin")

    fig.savefig(path, bbox_inches="tight")
    plt.close()


def plot_distribution_of_logits(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, axs = plt.subplots(figsize=(16, 10))

    data["proba_logit"] = np.log(data["y_pred"].values / (1 - data["y_pred"].values))
    sns.kdeplot(
        data[data["y_true"] == 0]["proba_logit"],
        fill=True,
        color="r",
        label="Class 0 logits",
        ax=axs,
    )
    sns.kdeplot(
        data[data["y_true"] == 1]["proba_logit"],
        fill=True,
        color="g",
        label="Class 1 logits",
        ax=axs,
    )
    axs.set_xlabel("Logits")
    axs.set_ylabel("Density")
    axs.set_title("Logits distribution of object predictions (by classes)")
    fig.savefig(path, bbox_inches="tight")
    plt.close()


def plot_pie_f1_metric(data, F1_thresh, path):
    tn, fp, fn, tp = confusion_matrix(data["y_true"], (data["y_pred"] > F1_thresh).astype(int)).ravel()
    (_, prec), (_, rec), (_, F1), (_, _) = precision_recall_fscore_support(
        data["y_true"], (data["y_pred"] > F1_thresh).astype(int)
    )

    sns.set(style="whitegrid", font_scale=1.5)
    fig, ax = plt.subplots(figsize=(20, 10), subplot_kw=dict(aspect="equal"))

    recipe = [
        "{} True Positives".format(tp),
        "{} False Positives".format(fp),
        "{} False Negatives".format(fn),
        "{} True Negatives".format(tn),
    ]

    wedges, texts = ax.pie([tp, fp, fn, tn], wedgeprops=dict(width=0.5), startangle=-40)

    bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
    kw = dict(
        arrowprops=dict(arrowstyle="-", color="k"),
        bbox=bbox_props,
        zorder=0,
        va="center",
    )

    for i, p in enumerate(wedges):
        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        ax.annotate(
            recipe[i], xy=(x, y), xytext=(1.35 * np.sign(x), 1.4 * y), horizontalalignment=horizontalalignment, **kw
        )

    ax.set_title(
        "Trained model: Precision = {:.2f}%, Recall = {:.2f}%, F1-Score = {:.2f}%".format(
            prec * 100, rec * 100, F1 * 100
        )
    )
    plt.savefig(path, bbox_inches="tight")
    plt.close()
    return prec, rec, F1


def f1_score_w_co(input_data, min_co=0.01, max_co=0.99, step=0.01):
    data = input_data.copy()
    data["y_pred"] = np.clip(np.ceil(data["y_pred"].values / step) * step, min_co, max_co)

    pos = data["y_true"].sum()
    neg = data["y_true"].shape[0] - pos

    grp = pd.DataFrame(data).groupby("y_pred")["y_true"].agg(["sum", "count"])
    grp.sort_index(inplace=True)

    grp["fp"] = grp["sum"].cumsum()
    grp["tp"] = pos - grp["fp"]
    grp["tn"] = (grp["count"] - grp["sum"]).cumsum()
    grp["fn"] = neg - grp["tn"]

    grp["pr"] = grp["tp"] / (grp["tp"] + grp["fp"])
    grp["rec"] = grp["tp"] / (grp["tp"] + grp["fn"])

    grp["f1_score"] = 2 * (grp["pr"] * grp["rec"]) / (grp["pr"] + grp["rec"])

    best_score = grp["f1_score"].max()
    best_co = grp.index.values[grp["f1_score"] == best_score].mean()

    # print((y_pred < best_co).mean())

    return best_score, best_co


def get_bins_table(data):
    bins_table = data.groupby("bin").agg({"y_true": [len, "mean"], "y_pred": ["min", "mean", "max"]}).reset_index()
    bins_table.columns = [
        "Bin number",
        "Amount of objects",
        "Mean target",
        "Min probability",
        "Average probability",
        "Max probability",
    ]
    return bins_table.to_html(index=False)


# Regression plots:


def plot_target_distribution_1(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, axs = plt.subplots(2, 1, figsize=(16, 20))

    sns.kdeplot(data["y_true"], fill=True, color="g", ax=axs[0])
    axs[0].set_xlabel("Target value")
    axs[0].set_ylabel("Density")
    axs[0].set_title("Target distribution (y_true)")

    sns.kdeplot(data["y_pred"], fill=True, color="r", ax=axs[1])
    axs[1].set_xlabel("Target value")
    axs[1].set_ylabel("Density")
    axs[1].set_title("Target distribution (y_pred)")

    fig.savefig(path, bbox_inches="tight")
    plt.close()


def plot_target_distribution_2(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, axs = plt.subplots(figsize=(16, 10))

    sns.kdeplot(data["y_true"], fill=True, color="g", label="y_true", ax=axs)
    sns.kdeplot(data["y_pred"], fill=True, color="r", label="y_pred", ax=axs)
    axs.set_xlabel("Target value")
    axs.set_ylabel("Density")
    axs.set_title("Target distribution")

    fig.savefig(path, bbox_inches="tight")
    plt.close()


def plot_target_distribution(data, path):
    data_pred = pd.DataFrame({"Target value": data["y_pred"]})
    data_pred["source"] = "y_pred"
    data_true = pd.DataFrame({"Target value": data["y_true"]})
    data_true["source"] = "y_true"
    data = pd.concat([data_pred, data_true], ignore_index=True)

    sns.set(style="whitegrid", font_scale=1.5)
    g = sns.displot(
        data,
        x="Target value",
        row="source",
        height=9,
        aspect=1.5,
        kde=True,
        color="m",
        facet_kws=dict(margin_titles=True),
    )
    g.fig.suptitle("Target distribution")
    g.fig.tight_layout()
    g.fig.subplots_adjust(top=0.95)

    g.fig.savefig(path, bbox_inches="tight")
    plt.close()


def plot_error_hist(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, ax = plt.subplots(figsize=(16, 10))

    sns.kdeplot(data["y_pred"] - data["y_true"], fill=True, color="m", ax=ax)
    ax.set_xlabel("Error = y_pred - y_true")
    ax.set_ylabel("Density")
    ax.set_title("Error histogram")

    fig.savefig(path, bbox_inches="tight")
    plt.close()


def plot_reg_scatter(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    g = sns.jointplot(
        x="y_pred",
        y="y_true",
        data=data,
        kind="reg",
        truncate=False,
        color="m",
        height=14,
    )
    g.fig.suptitle("Scatter plot")
    g.fig.tight_layout()
    g.fig.subplots_adjust(top=0.95)

    g.fig.savefig(path, bbox_inches="tight")
    plt.close()


# Multiclass plots:


def plot_confusion_matrix(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, ax = plt.subplots(figsize=(16, 12))

    cmat = confusion_matrix(data["y_true"], data["y_pred"], normalize="true")
    sns.heatmap(cmat, annot=True, linewidths=0.5, cmap="Purples", ax=ax)
    ax.set_xlabel("y_pred")
    ax.set_ylabel("y_true")
    ax.set_title("Confusion matrix")

    fig.savefig(path, bbox_inches="tight")
    plt.close()


# Feature importance


def plot_feature_importance(feat_imp, path, features_max=100):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, axs = plt.subplots(figsize=(16, features_max / 2.5))
    sns.barplot(x="Importance", y="Feature", data=feat_imp[:features_max], ax=axs, color="m")
    plt.savefig(path, bbox_inches="tight")
    plt.close()


def list2table(feature_list: list, html_params: dict = {}) -> str:
    """Creates HTML table with feature description from list of converts list of items.

    Args:
        feature_list: list of dictionaries with features' properties (e.g. name, length, stat. properties, etc.);
        html_params: extra parameters for pandas.DataFrame().to_html() function;

    Returns:
        String representation of HTML table.

    """
    default_html_params = {"index": False, "justify": "left"}
    default_html_params.update(html_params)
    if len(feature_list) == 0:
        return None
    else:
        return pd.DataFrame(feature_list).to_html(**default_html_params)


class ReportDeco:
    """
    Decorator to wrap :class:`~lightautoml.automl.base.AutoML` class to generate html report on ``fit_predict`` and ``predict``.

    Example:

        >>> report_automl = ReportDeco(output_path="output_path", report_file_name="report_file_name")(automl).
        >>> report_automl.fit_predict(train_data)
        >>> report_automl.predict(test_data)

    Report will be generated at output_path/report_file_name automatically.

    Warning:
         Do not use it just to inference (if you don"t need report), because:

            - It needs target variable to calc performance metrics.
            - It takes additional time to generate report.
            - Dump of decorated automl takes more memory to store.

    To get unwrapped fitted instance to pickle
    and inferecne access ``report_automl.model`` attribute.

    """

    @property
    def model(self):
        """Get unwrapped model.

        Returns:
            model.

        """
        return self._model

    @property
    def mapping(self):
        return self._model.reader.class_mapping

    @property
    def task(self):
        return self._model.reader.task._name

    def __init__(self, *args, **kwargs):
        """

        Note:
            Valid kwargs are:

                - output_path: Folder with report files.
                - report_file_name: Name of main report file.

        Args:
            *args: Arguments.
            **kwargs: Additional parameters.

        """
        if not kwargs:
            kwargs = {}

        # default params
        self.fi_params = {"method": "fast", "n_sample": 100_000}
        self.interpretation_params = {
            "top_n_features": 5,
            "top_n_categories": 10,
            "top_n_classes": 10,
            "n_bins": 30,
            "datetime_level": "year",
            "n_sample": 100_000,
        }

        fi_input_params = kwargs.get("fi_params", {})
        self.fi_params.update(fi_input_params)
        interpretation_input_params = kwargs.get("interpretation_params", {})
        self.interpretation_params.update(interpretation_input_params)
        self.interpretation = kwargs.get("interpretation", False)

        self.n_bins = kwargs.get("n_bins", 20)
        self.template_path = kwargs.get("template_path", os.path.join(base_dir, "lama_report_templates/"))
        self.output_path = kwargs.get("output_path", "lama_report/")
        self.report_file_name = kwargs.get("report_file_name", "lama_interactive_report.html")
        self.pdf_file_name = kwargs.get("pdf_file_name", None)

        if not os.path.exists(self.output_path):
            os.makedirs(self.output_path, exist_ok=True)

        self._base_template_path = "lama_base_template.html"
        self._model_section_path = "model_section.html"
        self._train_set_section_path = "train_set_section.html"
        self._results_section_path = "results_section.html"
        self._fi_section_path = "feature_importance_section.html"
        self._interpretation_section_path = "interpretation_section.html"
        self._interpretation_subsection_path = "interpretation_subsection.html"

        self._inference_section_path = {
            "binary": "binary_inference_section.html",
            "reg": "reg_inference_section.html",
            "multiclass": "multiclass_inference_section.html",
        }

        self.title = "LAMA report"
        if self.interpretation:
            self.sections_order = [
                "intro",
                "model",
                "train_set",
                "fi",
                "interpretation",
                "results",
            ]
            self._interpretation_top = []
        else:
            self.sections_order = ["intro", "model", "train_set", "fi", "results"]
        self._sections = {}
        self._sections["intro"] = "<p>This report was generated automatically.</p>"
        self._model_results = []

        self.generate_report()

    def __call__(self, model):
        self._model = model

        # add informataion to report
        self._model_name = model.__class__.__name__
        self._model_parameters = json2html.convert(extract_params(model))
        self._model_summary = None

        self._sections = {}
        self._sections["intro"] = "<p>This report was generated automatically.</p>"
        self._model_results = []
        self._n_test_sample = 0

        self._generate_model_section()
        self.generate_report()
        return self

    def _binary_classification_details(self, data):
        self._inference_content["sample_bins_table"] = get_bins_table(data)
        prec, rec, F1 = plot_pie_f1_metric(
            data,
            self._F1_thresh,
            path=os.path.join(self.output_path, self._inference_content["pie_f1_metric"]),
        )
        auc_score = plot_roc_curve_image(
            data,
            path=os.path.join(self.output_path, self._inference_content["roc_curve"]),
        )
        plot_pr_curve_image(
            data,
            path=os.path.join(self.output_path, self._inference_content["pr_curve"]),
        )
        plot_preds_distribution_by_bins(
            data,
            path=os.path.join(self.output_path, self._inference_content["preds_distribution_by_bins"]),
        )
        plot_distribution_of_logits(
            data,
            path=os.path.join(self.output_path, self._inference_content["distribution_of_logits"]),
        )
        return auc_score, prec, rec, F1

    def _regression_details(self, data):
        # graphics
        plot_target_distribution(
            data,
            path=os.path.join(self.output_path, self._inference_content["target_distribution"]),
        )
        plot_error_hist(
            data,
            path=os.path.join(self.output_path, self._inference_content["error_hist"]),
        )
        plot_reg_scatter(
            data,
            path=os.path.join(self.output_path, self._inference_content["scatter_plot"]),
        )
        # metrics
        mean_ae = mean_absolute_error(data["y_true"], data["y_pred"])
        median_ae = median_absolute_error(data["y_true"], data["y_pred"])
        mse = mean_squared_error(data["y_true"], data["y_pred"])
        r2 = r2_score(data["y_true"], data["y_pred"])
        evs = explained_variance_score(data["y_true"], data["y_pred"])
        return mean_ae, median_ae, mse, r2, evs

    def _multiclass_details(self, data):
        y_true = data["y_true"]
        y_pred = data["y_pred"]
        # precision
        p_micro = precision_score(y_true, y_pred, average="micro")
        p_macro = precision_score(y_true, y_pred, average="macro")
        p_weighted = precision_score(y_true, y_pred, average="weighted")
        # recall
        r_micro = recall_score(y_true, y_pred, average="micro")
        r_macro = recall_score(y_true, y_pred, average="macro")
        r_weighted = recall_score(y_true, y_pred, average="weighted")
        # f1-score
        f_micro = f1_score(y_true, y_pred, average="micro")
        f_macro = f1_score(y_true, y_pred, average="macro")
        f_weighted = f1_score(y_true, y_pred, average="weighted")

        # classification report for features
        if self.mapping:
            classes = sorted(self.mapping, key=self.mapping.get)
        else:
            classes = np.arange(self._N_classes)
        p, r, f, s = precision_recall_fscore_support(y_true, y_pred)
        cls_report = pd.DataFrame(
            {
                "Class name": classes,
                "Precision": p,
                "Recall": r,
                "F1-score": f,
                "Support": s,
            }
        )
        self._inference_content["classification_report"] = cls_report.to_html(
            index=False, float_format="{:.4f}".format, justify="left"
        )

        plot_confusion_matrix(
            data,
            path=os.path.join(self.output_path, self._inference_content["confusion_matrix"]),
        )

        return [
            p_micro,
            p_macro,
            p_weighted,
            r_micro,
            r_macro,
            r_weighted,
            f_micro,
            f_macro,
            f_weighted,
        ]

    def _collect_data(self, preds, sample):
        data = pd.DataFrame({"y_true": sample[self._target].values})
        if self.task in "multiclass":
            if self.mapping is not None:
                data["y_true"] = np.array([self.mapping[y] for y in data["y_true"].values])
            data["y_pred"] = preds._data.argmax(axis=1)
            data = data[~np.isnan(preds._data).any(axis=1)]
        else:
            data["y_pred"] = preds._data[:, 0]
            data.sort_values("y_pred", ascending=False, inplace=True)
            data["bin"] = (np.arange(data.shape[0]) / data.shape[0] * self.n_bins).astype(int)
            data = data[~data["y_pred"].isnull()]
        return data

    def fit_predict(self, *args, **kwargs):
        """Wrapped ``automl.fit_predict`` method.

        Valid args, kwargs are the same as wrapped automl.

        Args:
            *args: Arguments.
            **kwargs: Additional parameters.

        Returns:
            OOF predictions.

        """
        # TODO: parameters parsing in general case

        preds = self._model.fit_predict(*args, **kwargs)
        train_data = kwargs["train_data"] if "train_data" in kwargs else args[0]
        input_roles = kwargs["roles"] if "roles" in kwargs else args[1]
        self._target = input_roles["target"]
        valid_data = kwargs.get("valid_data", None)
        if valid_data is None:
            data = self._collect_data(preds, train_data)
        else:
            data = self._collect_data(preds, valid_data)
        self._inference_content = {}
        if self.task == "binary":
            # filling for html
            self._inference_content = {}
            self._inference_content["roc_curve"] = "valid_roc_curve.png"
            self._inference_content["pr_curve"] = "valid_pr_curve.png"
            self._inference_content["pie_f1_metric"] = "valid_pie_f1_metric.png"
            self._inference_content["preds_distribution_by_bins"] = "valid_preds_distribution_by_bins.png"
            self._inference_content["distribution_of_logits"] = "valid_distribution_of_logits.png"
            # graphics and metrics
            _, self._F1_thresh = f1_score_w_co(data)
            auc_score, prec, rec, F1 = self._binary_classification_details(data)
            # update model section
            evaluation_parameters = ["AUC-score", "Precision", "Recall", "F1-score"]
            self._model_summary = pd.DataFrame(
                {
                    "Evaluation parameter": evaluation_parameters,
                    "Validation sample": [auc_score, prec, rec, F1],
                }
            )
        elif self.task == "reg":
            # filling for html
            self._inference_content["target_distribution"] = "valid_target_distribution.png"
            self._inference_content["error_hist"] = "valid_error_hist.png"
            self._inference_content["scatter_plot"] = "valid_scatter_plot.png"
            # graphics and metrics
            mean_ae, median_ae, mse, r2, evs = self._regression_details(data)
            # model section
            evaluation_parameters = [
                "Mean absolute error",
                "Median absolute error",
                "Mean squared error",
                "R^2 (coefficient of determination)",
                "Explained variance",
            ]
            self._model_summary = pd.DataFrame(
                {
                    "Evaluation parameter": evaluation_parameters,
                    "Validation sample": [mean_ae, median_ae, mse, r2, evs],
                }
            )
        elif self.task == "multiclass":
            self._N_classes = len(train_data[self._target].drop_duplicates())
            self._inference_content["confusion_matrix"] = "valid_confusion_matrix.png"

            index_names = np.array([["Precision", "Recall", "F1-score"], ["micro", "macro", "weighted"]])
            index = pd.MultiIndex.from_product(index_names, names=["Evaluation metric", "Average"])

            summary = self._multiclass_details(data)
            self._model_summary = pd.DataFrame({"Validation sample": summary}, index=index)

        self._inference_content["title"] = "Results on validation sample"

        self._generate_model_section()

        # generate train data section
        self._train_data_overview = self._data_genenal_info(train_data)
        self._describe_roles(train_data)
        self._describe_dropped_features(train_data)
        self._generate_train_set_section()
        # generate fit_predict section
        self._generate_inference_section()
        # generate feature importance and interpretation sections
        self._generate_fi_section(valid_data)
        if self.interpretation:
            self._generate_interpretation_section(valid_data)

        self.generate_report()
        return preds

    def predict(self, *args, **kwargs):
        """Wrapped automl.predict method.

        Valid args, kwargs are the same as wrapped automl.

        Args:
            *args: arguments.
            **kwargs: additional parameters.

        Returns:
            predictions.

        """
        self._n_test_sample += 1
        # get predictions
        test_preds = self._model.predict(*args, **kwargs)

        test_data = kwargs["test"] if "test" in kwargs else args[0]
        data = self._collect_data(test_preds, test_data)

        if self.task == "binary":
            # filling for html
            self._inference_content = {}
            self._inference_content["roc_curve"] = "test_roc_curve_{}.png".format(self._n_test_sample)
            self._inference_content["pr_curve"] = "test_pr_curve_{}.png".format(self._n_test_sample)
            self._inference_content["pie_f1_metric"] = "test_pie_f1_metric_{}.png".format(self._n_test_sample)
            self._inference_content["bins_preds"] = "test_bins_preds_{}.png".format(self._n_test_sample)
            self._inference_content["preds_distribution_by_bins"] = "test_preds_distribution_by_bins_{}.png".format(
                self._n_test_sample
            )
            self._inference_content["distribution_of_logits"] = "test_distribution_of_logits_{}.png".format(
                self._n_test_sample
            )
            # graphics and metrics
            auc_score, prec, rec, F1 = self._binary_classification_details(data)

            if self._n_test_sample >= 2:
                self._model_summary["Test sample {}".format(self._n_test_sample)] = [
                    auc_score,
                    prec,
                    rec,
                    F1,
                ]
            else:
                self._model_summary["Test sample"] = [auc_score, prec, rec, F1]

        elif self.task == "reg":
            # filling for html
            self._inference_content = {}
            self._inference_content["target_distribution"] = "test_target_distribution_{}.png".format(
                self._n_test_sample
            )
            self._inference_content["error_hist"] = "test_error_hist_{}.png".format(self._n_test_sample)
            self._inference_content["scatter_plot"] = "test_scatter_plot_{}.png".format(self._n_test_sample)
            # graphics
            mean_ae, median_ae, mse, r2, evs = self._regression_details(data)
            # update model section
            if self._n_test_sample >= 2:
                self._model_summary["Test sample {}".format(self._n_test_sample)] = [
                    mean_ae,
                    median_ae,
                    mse,
                    r2,
                    evs,
                ]
            else:
                self._model_summary["Test sample"] = [mean_ae, median_ae, mse, r2, evs]

        elif self.task == "multiclass":
            self._inference_content["confusion_matrix"] = "test_confusion_matrix_{}.png".format(self._n_test_sample)
            test_summary = self._multiclass_details(data)
            if self._n_test_sample >= 2:
                self._model_summary["Test sample {}".format(self._n_test_sample)] = test_summary
            else:
                self._model_summary["Test sample"] = test_summary

        # layout depends on number of test samples
        if self._n_test_sample >= 2:
            self._inference_content["title"] = "Results on test sample {}".format(self._n_test_sample)

        else:
            self._inference_content["title"] = "Results on test sample"

        # update model section
        self._generate_model_section()
        # generate predict section
        self._generate_inference_section()

        self.generate_report()
        return test_preds

    def _generate_fi_section(self, valid_data):
        if (
            self.fi_params["method"] == "accurate"
            and valid_data is not None
            and valid_data.shape[0] > self.fi_params["n_sample"]
        ):
            valid_data = valid_data.sample(n=self.fi_params["n_sample"])
            print(
                "valid_data was sampled for feature importance calculation: n_sample = {}".format(
                    self.fi_params["n_sample"]
                )
            )

        if self.fi_params["method"] == "accurate" and valid_data is None:
            # raise ValueError("You must set valid_data with accurate feature importance method")
            self.fi_params["method"] = "fast"
            warnings.warn(
                "You must set valid_data with 'accurate' feature importance method. Changed to 'fast' automatically."
            )

        self.feat_imp = self._model.get_feature_scores(
            calc_method=self.fi_params["method"], data=valid_data, silent=False
        )
        if self.feat_imp is None:
            fi_path = None
        else:
            fi_path = "feature_importance.png"
            plot_feature_importance(self.feat_imp, path=os.path.join(self.output_path, fi_path))
        # add to _sections
        fi_content = {
            "fi_method": self.fi_params["method"],
            "feature_importance": fi_path,
        }
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        fi_section = env.get_template(self._fi_section_path).render(fi_content)
        self._sections["fi"] = fi_section

    def _generate_interpretation_content(self, test_data):
        self._interpretation_content = {}
        if test_data is None:
            self._interpretation_content["interpretation_top"] = None
            return
        if self.feat_imp is None:
            interpretation_feat_list = list(self._model.reader._roles.keys())[
                : self.interpretation_params["top_n_features"]
            ]
        else:
            interpretation_feat_list = self.feat_imp["Feature"].values[: self.interpretation_params["top_n_features"]]
        for feature_name in interpretation_feat_list:
            interpretaton_subsection = {}
            interpretaton_subsection["feature_name"] = feature_name
            interpretaton_subsection["feature_interpretation_plot"] = feature_name + "_interpretation.png"
            self._plot_pdp(
                test_data,
                feature_name,
                path=os.path.join(
                    self.output_path,
                    interpretaton_subsection["feature_interpretation_plot"],
                ),
            )
            env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
            interpretation_subsection = env.get_template(self._interpretation_subsection_path).render(
                interpretaton_subsection
            )
            self._interpretation_top.append(interpretation_subsection)
            print(f"Interpretation info for {feature_name} appended")
        self._interpretation_content["interpretation_top"] = self._interpretation_top

    def _generate_interpretation_section(self, test_data):
        if test_data is not None and test_data.shape[0] > self.interpretation_params["n_sample"]:
            test_data = test_data.sample(n=self.interpretation_params["n_sample"])
        self._generate_interpretation_content(test_data)
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        interpretation_section = env.get_template(self._interpretation_section_path).render(
            self._interpretation_content
        )
        self._sections["interpretation"] = interpretation_section

    def _plot_pdp(self, test_data, feature_name, path):
        feature_role = self._model.reader._roles[feature_name].name
        # I. Count interpretation
        print("Calculating interpretation for {}:".format(feature_name))
        grid, ys, counts = self._model.get_individual_pdp(
            test_data=test_data,
            feature_name=feature_name,
            n_bins=self.interpretation_params["n_bins"],
            top_n_categories=self.interpretation_params["top_n_categories"],
            datetime_level=self.interpretation_params["datetime_level"],
        )
        # II. Plot pdp
        sns.set(style="whitegrid", font_scale=1.5)
        fig, axs = plt.subplots(2, 1, figsize=(16, 12), gridspec_kw={"height_ratios": [3, 1]})
        axs[0].set_title("PDP plot: " + feature_name)
        n_classes = ys[0].shape[1]
        if n_classes == 1:
            data = pd.concat(
                [pd.DataFrame({"x": grid[i], "y": ys[i].ravel()}) for i, _ in enumerate(grid)]
            ).reset_index(drop=True)
            if feature_role in ["Numeric", "Datetime"]:
                g0 = sns.lineplot(data=data, x="x", y="y", ax=axs[0], color="m")
            else:
                g0 = sns.boxplot(data=data, x="x", y="y", ax=axs[0], showfliers=False, color="m")
        else:
            if self.mapping:
                classes = sorted(self.mapping, key=self.mapping.get)[: self.interpretation_params["top_n_classes"]]
            else:
                classes = np.arange(min(n_classes, self.interpretation_params["top_n_classes"]))
            data = pd.concat(
                [
                    pd.DataFrame({"x": grid[i], "y": ys[i][:, k], "class": name})
                    for i, _ in enumerate(grid)
                    for k, name in enumerate(classes)
                ]
            ).reset_index(drop=True)
            if self._model.reader._roles[feature_name].name in ["Numeric", "Datetime"]:
                g0 = sns.lineplot(data=data, x="x", y="y", hue="class", ax=axs[0])
            else:
                g0 = sns.boxplot(data=data, x="x", y="y", hue="class", ax=axs[0], showfliers=False)
        g0.set(ylabel="y_pred")
        # III. Plot distribution
        counts = np.array(counts) / sum(counts)
        if feature_role == "Numeric":
            g0.set(xlabel="feature value")
            g1 = sns.histplot(test_data[feature_name], kde=True, color="gray", ax=axs[1])
        elif feature_role == "Category":
            g0.set(xlabel=None)
            axs[0].set_xticklabels(grid, rotation=90)
            g1 = sns.barplot(x=grid, y=counts, ax=axs[1], color="gray")
        else:
            g0.set(xlabel=self.interpretation_params["datetime_level"])
            g1 = sns.barplot(x=grid, y=counts, ax=axs[1], color="gray")
        g1.set(xlabel=None)
        g1.set(ylabel="Frequency")
        g1.set(xticklabels=[])
        # IV. Save picture
        plt.tight_layout()
        fig.savefig(path, bbox_inches="tight")
        plt.close()

    def _data_genenal_info(self, data):
        general_info = pd.DataFrame(columns=["Parameter", "Value"])
        general_info.loc[0] = ("Number of records", data.shape[0])
        general_info.loc[1] = ("Total number of features", data.shape[1])
        general_info.loc[2] = ("Used features", len(self._model.reader._used_features))
        general_info.loc[3] = (
            "Dropped features",
            len(self._model.reader._dropped_features),
        )
        # general_info.loc[4] = ("Number of positive cases", np.sum(data[self._target] == 1))
        # general_info.loc[5] = ("Number of negative cases", np.sum(data[self._target] == 0))
        return general_info.to_html(index=False, justify="left")

    def _describe_roles(self, train_data):

        # detect feature roles
        roles = self._model.reader._roles
        numerical_features = [feat_name for feat_name in roles if roles[feat_name].name == "Numeric"]
        categorical_features = [feat_name for feat_name in roles if roles[feat_name].name == "Category"]
        datetime_features = [feat_name for feat_name in roles if roles[feat_name].name == "Datetime"]
        text_features = [feat_name for feat_name in roles if roles[feat_name].name == "Text"]

        # numerical roles
        numerical_features_df = []
        for feature_name in numerical_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(train_data[feature_name].isna().sum() / train_data.shape[0])
            # check if column dtype is bool
            if train_data[feature_name].dtype == bool:
                values = train_data[feature_name].astype(float).dropna().values
            else:
                values = train_data[feature_name].dropna().values
            item["min"] = np.min(values)
            item["quantile_25"] = np.quantile(values, 0.25)
            item["average"] = np.mean(values)
            item["median"] = np.median(values)
            item["quantile_75"] = np.quantile(values, 0.75)
            item["max"] = np.max(values)
            numerical_features_df.append(item)
        self._numerical_features_table = list2table(numerical_features_df, {"float_format": "{:.2f}".format})

        # categorical roles
        categorical_features_df = []
        for feature_name in categorical_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(train_data[feature_name].isna().sum() / train_data.shape[0])
            value_counts = train_data[feature_name].value_counts(normalize=True)
            values = value_counts.index.values
            counts = value_counts.values
            item["Number of unique values"] = len(counts)
            item["Most frequent value"] = values[0]
            item["Occurrence of most frequent"] = "{:.1f}%".format(100 * counts[0])
            item["Least frequent value"] = values[-1]
            item["Occurrence of least frequent"] = "{:.1f}%".format(100 * counts[-1])
            categorical_features_df.append(item)
        self._categorical_features_table = list2table(categorical_features_df)

        # datetime roles
        datetime_features_df = []
        for feature_name in datetime_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(train_data[feature_name].isna().sum() / train_data.shape[0])
            values = train_data[feature_name].dropna().values
            item["min"] = np.min(values)
            item["max"] = np.max(values)
            item["base_date"] = self._model.reader._roles[feature_name].base_date
            datetime_features_df.append(item)
        self._datetime_features_table = list2table(datetime_features_df)

        # text roles
        text_features_df = []
        for feature_name in text_features:
            item = {"Feature name": feature_name}
            feature_length = train_data[feature_name].str.len()
            item["Amount of empty records"] = (feature_length == 0).sum(axis=0)
            item["Length of the shortest sentence"] = feature_length.min()
            item["Length of the longest sentence"] = feature_length.max()
            text_features_df.append(item)
        self._text_features_table = list2table(text_features_df)

    def _describe_dropped_features(self, train_data):
        self._max_nan_rate = self._model.reader.max_nan_rate
        self._max_constant_rate = self._model.reader.max_constant_rate
        self._features_dropped_list = self._model.reader._dropped_features
        # dropped features table
        dropped_list = [col for col in self._features_dropped_list if col != self._target]
        if dropped_list == []:
            self._dropped_features_table = None
        else:
            dropped_nan_ratio = train_data[dropped_list].isna().sum() / train_data.shape[0]
            dropped_most_occured = pd.Series(np.nan, index=dropped_list)
            for col in dropped_list:
                col_most_occured = train_data[col].value_counts(normalize=True).values
                if len(col_most_occured) > 0:
                    dropped_most_occured[col] = col_most_occured[0]
            dropped_features_table = pd.DataFrame(
                {"nan_rate": dropped_nan_ratio, "constant_rate": dropped_most_occured}
            )
            self._dropped_features_table = (
                dropped_features_table.reset_index()
                .rename(columns={"index": "Название переменной"})
                .to_html(index=False, justify="left")
            )

    def _generate_model_section(self):
        model_summary = None
        if self._model_summary is not None:
            model_summary = self._model_summary.to_html(
                index=self.task == "multiclass",
                justify="left",
                float_format="{:.4f}".format,
            )

        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        model_section = env.get_template(self._model_section_path).render(
            model_name=self._model_name,
            model_parameters=self._model_parameters,
            model_summary=model_summary,
        )
        self._sections["model"] = model_section

    def _generate_train_set_section(self):
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        train_set_section = env.get_template(self._train_set_section_path).render(
            train_data_overview=self._train_data_overview,
            numerical_features_table=self._numerical_features_table,
            categorical_features_table=self._categorical_features_table,
            datetime_features_table=self._datetime_features_table,
            text_features_table=self._text_features_table,
            target=self._target,
            max_nan_rate=self._max_nan_rate,
            max_constant_rate=self._max_constant_rate,
            dropped_features_table=self._dropped_features_table,
        )
        self._sections["train_set"] = train_set_section

    def _generate_inference_section(self):
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        inference_section = env.get_template(self._inference_section_path[self.task]).render(self._inference_content)
        self._model_results.append(inference_section)

    def _generate_results_section(self):
        if self._model_results:
            env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
            results_section = env.get_template(self._results_section_path).render(model_results=self._model_results)
            self._sections["results"] = results_section

    def generate_report(self):
        # collection sections
        self._generate_results_section()
        sections_list = []
        for sec_name in self.sections_order:
            if sec_name in self._sections:
                sections_list.append(self._sections[sec_name])
        # put sections inside
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        report = env.get_template(self._base_template_path).render(
            title=self.title, sections=sections_list, pdf=self.pdf_file_name
        )

        with open(os.path.join(self.output_path, self.report_file_name), "w", encoding="utf-8") as f:
            f.write(report)

        if self.pdf_file_name:
            try:
                from weasyprint import HTML

                HTML(string=report, base_url=self.output_path).write_pdf(
                    os.path.join(self.output_path, self.pdf_file_name)
                )
            except ModuleNotFoundError:
                print("Can't generate PDF report: check manual for installing pdf extras.")


_config_name_desc_dict = {
    "conf_0_sel_type_0.yml": "Preset 0: No feature selection",
    "conf_1_sel_type_1.yml": "Preset 1: Cutoff feature selection",
    "conf_2_select_mode_1_no_typ.yml": "Preset 2: Cutoff feature selection, no advanced typisation",
    "conf_3_sel_type_1_no_inter_lgbm.yml": "Preset 3: Cutoff feature selection, no GBM categories intersections",
    "conf_4_sel_type_0_no_int.yml": "Preset 4: No feature selection, no Linear categories intersections, no GBM categories intersections",
    "conf_5_sel_type_1_tuning_full.yml": "Preset 5: Cutoff feature selection, accurate parameters tuner",
    "conf_6_sel_type_1_tuning_full_no_int_lgbm.yml": "Preset 6: Cutoff feature selection, accurate parameters tuner, no GBM categories intersections",
}


class ReportDecoUtilized(ReportDeco):
    """
    Special report wrapper for :class:`~lightautoml.automl.presets.tabular_presets.TabularUtilizedAutoML`.
    Usage case is the same as main
    :class:`~lightautoml.report.report_deco.ReportDeco` class.
    It generates same report as :class:`~lightautoml.report.report_deco.ReportDeco` ,
    but with info about each used preset.

    Difference:

        - the model parameters section appears only after fit_predict is called.
          report_automl.__call__ and report_automl.fit_predict are changed.
        - report_automl._generate_model_section and report_automl._generate_train_set_section obtain
          the list of subsections of Model and Data overview and concatenate in one section.
        - report_automl._generate_preset_sections and report_automl._generate_data_sections
          are the new functions that generate Model overview and Data overview section for each preset individually.
        - report_automl._data_genenal_info, report_automl._describe_roles,
          report_automl._describe_dropped_features get on input additional reader
          parameter. Each preset has its own reader configuration, so train data
          summary is generated for each reader individually.
        - new html templates are intoduced.

    """

    @property
    def task(self):
        return self._model.task.name

    @property
    def mapping(self):
        return self._model.outer_pipes[0].ml_algos[0].models[0][0].reader.class_mapping

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._fi_section_path = "feature_importance_utillized_section.html"
        self._model_section_path = "model_section_utilized.html"
        self._preset_section_path = "preset_section.html"
        self._preset_sections = None
        self._preset_data_sections_path = "utilized_data_subsections.html"
        self._data_sections = None
        self._train_set_section_path = "train_set_section_utilized.html"

        self._roles_df = pd.DataFrame()
        self._pred_formula = None

    def __call__(self, model):
        self._model = model

        # add informataion to report
        self._model_name = model.__class__.__name__
        self._model_summary = None

        self._sections = {}
        self._sections["intro"] = "<p>This report was generated automatically.</p>"
        self._model_results = []
        self._n_test_sample = 0

        self._generate_model_section()
        self.generate_report()
        return self

    def fit_predict(self, *args, **kwargs):
        """Wrapped :meth:`TimeUtilization.fit_predict` method.

        Valid args, kwargs are the same as wrapped automl.

        Args:
            *args: Arguments.
            **kwargs: Additional parameters.

        Returns:
            OOF predictions.

        """
        # TODO: parameters parsing in general case

        preds = self._model.fit_predict(*args, **kwargs)

        # Generate sections describing presets used in `fit_predict`
        self._generate_preset_sections()

        train_data = kwargs["train_data"] if "train_data" in kwargs else args[0]
        input_roles = kwargs["roles"] if "roles" in kwargs else args[1]
        self._target = input_roles["target"]
        valid_data = kwargs.get("valid_data", None)
        if valid_data is None:
            data = self._collect_data(preds, train_data)
        else:
            data = self._collect_data(preds, valid_data)
        self._inference_content = {}
        if self.task == "binary":
            # filling for html
            self._inference_content = {}
            self._inference_content["roc_curve"] = "valid_roc_curve.png"
            self._inference_content["pr_curve"] = "valid_pr_curve.png"
            self._inference_content["pie_f1_metric"] = "valid_pie_f1_metric.png"
            self._inference_content["preds_distribution_by_bins"] = "valid_preds_distribution_by_bins.png"
            self._inference_content["distribution_of_logits"] = "valid_distribution_of_logits.png"
            # graphics and metrics
            _, self._F1_thresh = f1_score_w_co(data)
            auc_score, prec, rec, F1 = self._binary_classification_details(data)
            # update model section
            evaluation_parameters = ["AUC-score", "Precision", "Recall", "F1-score"]
            self._model_summary = pd.DataFrame(
                {
                    "Evaluation parameter": evaluation_parameters,
                    "Validation sample": [auc_score, prec, rec, F1],
                }
            )
        elif self.task == "reg":
            # filling for html
            self._inference_content["target_distribution"] = "valid_target_distribution.png"
            self._inference_content["error_hist"] = "valid_error_hist.png"
            self._inference_content["scatter_plot"] = "valid_scatter_plot.png"
            # graphics and metrics
            mean_ae, median_ae, mse, r2, evs = self._regression_details(data)
            # model section
            evaluation_parameters = [
                "Mean absolute error",
                "Median absolute error",
                "Mean squared error",
                "R^2 (coefficient of determination)",
                "Explained variance",
            ]
            self._model_summary = pd.DataFrame(
                {
                    "Evaluation parameter": evaluation_parameters,
                    "Validation sample": [mean_ae, median_ae, mse, r2, evs],
                }
            )
        elif self.task == "multiclass":
            self._N_classes = len(train_data[self._target].drop_duplicates())
            self._inference_content["confusion_matrix"] = "valid_confusion_matrix.png"

            index_names = np.array([["Precision", "Recall", "F1-score"], ["micro", "macro", "weighted"]])
            index = pd.MultiIndex.from_product(index_names, names=["Evaluation metric", "Average"])

            summary = self._multiclass_details(data)
            self._model_summary = pd.DataFrame({"Validation sample": summary}, index=index)

        self._inference_content["title"] = "Results on validation sample"

        # wrap text in <pre> tag to display \n, \t
        self._pred_formula = "<pre>" + self._model.create_model_str_desc() + "</pre>"

        self._generate_model_section()

        # generate train data section
        self._data_sections = []
        for model in self._model.outer_pipes:
            preset_name = model.ml_algos[0].models[0][0].config_path.split("/")[-1].split(".")[0]
            reader = model.ml_algos[0].models[0][0].reader
            self._train_data_overview = self._data_genenal_info(train_data, reader)
            self._describe_roles(train_data, reader, preset_name)
            self._describe_dropped_features(train_data, reader, model.ml_algos[0].models[0][0], preset_name)
            train_set_section = self._generate_data_sections(preset_name)
            self._data_sections.append(train_set_section)

        self._roles_df = self._roles_df.T.to_html(justify="left")
        self._generate_train_set_section()

        # generate fit_predict section
        self._generate_inference_section()
        # generate feature importance and interpretation sections
        self._generate_fi_section(valid_data)
        if self.interpretation:
            self._generate_interpretation_section(valid_data)

        self.generate_report()
        return preds

    def _data_genenal_info(self, data, reader):
        general_info = pd.DataFrame(columns=["Parameter", "Value"])
        general_info.loc[0] = ("Number of records", data.shape[0])
        general_info.loc[1] = ("Total number of features", data.shape[1])
        general_info.loc[2] = ("Used features", len(reader._used_features))
        general_info.loc[3] = (
            "Dropped features",
            len(reader._dropped_features),
        )
        # general_info.loc[4] = ("Number of positive cases", np.sum(data[self._target] == 1))
        # general_info.loc[5] = ("Number of negative cases", np.sum(data[self._target] == 0))
        return general_info.to_html(index=False, justify="left")

    def _describe_roles(self, train_data, reader, preset_name):

        # detect feature roles
        roles = reader._roles
        numerical_features = [feat_name for feat_name in roles if roles[feat_name].name == "Numeric"]
        categorical_features = [feat_name for feat_name in roles if roles[feat_name].name == "Category"]
        datetime_features = [feat_name for feat_name in roles if roles[feat_name].name == "Datetime"]
        text_features = [feat_name for feat_name in roles if roles[feat_name].name == "Text"]

        # filling data frame with roles description
        self._roles_df.loc[preset_name, numerical_features] = ["N" for _ in range(len(numerical_features))]
        self._roles_df.loc[preset_name, categorical_features] = ["C" for _ in range(len(categorical_features))]
        self._roles_df.loc[preset_name, datetime_features] = ["D" for _ in range(len(datetime_features))]
        self._roles_df.loc[preset_name, text_features] = ["T" for _ in range(len(text_features))]

        # numerical roles
        numerical_features_df = []
        for feature_name in numerical_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(train_data[feature_name].isna().sum() / train_data.shape[0])
            # check if column dtype is bool
            if train_data[feature_name].dtype == bool:
                values = train_data[feature_name].astype(float).dropna().values
            else:
                values = train_data[feature_name].dropna().values
            item["min"] = np.min(values)
            item["quantile_25"] = np.quantile(values, 0.25)
            item["average"] = np.mean(values)
            item["median"] = np.median(values)
            item["quantile_75"] = np.quantile(values, 0.75)
            item["max"] = np.max(values)
            numerical_features_df.append(item)
        self._numerical_features_table = list2table(numerical_features_df, {"float_format": "{:.2f}".format})

        # categorical roles
        categorical_features_df = []
        for feature_name in categorical_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(train_data[feature_name].isna().sum() / train_data.shape[0])
            value_counts = train_data[feature_name].value_counts(normalize=True)
            values = value_counts.index.values
            counts = value_counts.values
            item["Number of unique values"] = len(counts)
            item["Most frequent value"] = values[0]
            item["Occurrence of most frequent"] = "{:.1f}%".format(100 * counts[0])
            item["Least frequent value"] = values[-1]
            item["Occurrence of least frequent"] = "{:.1f}%".format(100 * counts[-1])
            categorical_features_df.append(item)
        self._categorical_features_table = list2table(categorical_features_df)

        # datetime roles
        datetime_features_df = []
        for feature_name in datetime_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(train_data[feature_name].isna().sum() / train_data.shape[0])
            values = train_data[feature_name].dropna().values
            item["min"] = np.min(values)
            item["max"] = np.max(values)
            item["base_date"] = reader._roles[feature_name].base_date
            datetime_features_df.append(item)
        self._datetime_features_table = list2table(datetime_features_df)

        # text roles
        text_features_df = []
        for feature_name in text_features:
            item = {"Feature name": feature_name}
            feature_length = train_data[feature_name].str.len()
            item["Amount of empty records"] = (feature_length == 0).sum(axis=0)
            item["Length of the shortest sentence"] = feature_length.min()
            item["Length of the longest sentence"] = feature_length.max()
            text_features_df.append(item)
        self._text_features_table = list2table(text_features_df)

    def _describe_dropped_features(self, train_data, reader, model, preset_name):
        self._max_nan_rate = reader.max_nan_rate
        self._max_constant_rate = reader.max_constant_rate
        self._features_dropped_list = reader._dropped_features

        selector_dropped_features = list(set(list(reader._roles.keys())) - set(model.collect_used_feats()))

        # filling data frame with roles description
        self._roles_df.loc[preset_name, reader._dropped_features] = ["-" for _ in range(len(reader._dropped_features))]
        self._roles_df.loc[preset_name, selector_dropped_features] = [
            "-" for _ in range(len(selector_dropped_features))
        ]
        # dropped features table
        dropped_list = [col for col in self._features_dropped_list if col != self._target]
        if dropped_list == []:
            self._dropped_features_table = None
        else:
            dropped_nan_ratio = train_data[dropped_list].isna().sum() / train_data.shape[0]
            dropped_most_occured = pd.Series(np.nan, index=dropped_list)
            for col in dropped_list:
                col_most_occured = train_data[col].value_counts(normalize=True).values
                if len(col_most_occured) > 0:
                    dropped_most_occured[col] = col_most_occured[0]
            dropped_features_table = pd.DataFrame(
                {"nan_rate": dropped_nan_ratio, "constant_rate": dropped_most_occured}
            )
            self._dropped_features_table = (
                dropped_features_table.reset_index()
                .rename(columns={"index": "Название переменной"})
                .to_html(index=False, justify="left")
            )

    def _generate_model_section(self):
        model_summary = None
        if self._model_summary is not None:
            model_summary = self._model_summary.to_html(
                index=self.task == "multiclass",
                justify="left",
                float_format="{:.4f}".format,
            )

        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        model_section = env.get_template(self._model_section_path).render(
            model_name=self._model_name,
            model_presets=self._preset_sections,
            model_summary=model_summary,
            pred_formula=self._pred_formula,
        )
        self._sections["model"] = model_section

    def _generate_train_set_section(self):
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        train_set_section = env.get_template(self._train_set_section_path).render(
            data_sections=self._data_sections, roles_table=self._roles_df
        )
        self._sections["train_set"] = train_set_section

    def _generate_preset_sections(self):

        self._preset_sections = []
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        for model in self._model.outer_pipes:
            preset_name = model.ml_algos[0].models[0][0].config_path.split("/")[-1]
            preset_desc = _config_name_desc_dict.get(preset_name, None)
            if preset_desc is not None:
                preset_name = "{0} ({1})".format(preset_desc, preset_name)

            model_parameters = json2html.convert(extract_params(model.ml_algos[0].models[0][0]))
            preset_section = env.get_template(self._preset_section_path).render(
                preset_name=preset_name, model_parameters=model_parameters
            )

            self._preset_sections.append(preset_section)

    def _generate_data_sections(self, preset_name):
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        train_set_section = env.get_template(self._preset_data_sections_path).render(
            train_data_overview=self._train_data_overview,
            numerical_features_table=self._numerical_features_table,
            categorical_features_table=self._categorical_features_table,
            datetime_features_table=self._datetime_features_table,
            text_features_table=self._text_features_table,
            target=self._target,
            max_nan_rate=self._max_nan_rate,
            max_constant_rate=self._max_constant_rate,
            dropped_features_table=self._dropped_features_table,
            preset_name=preset_name,
        )
        return train_set_section


_default_wb_report_params = {
    "automl_date_column": "",
    "report_name": "autowoe_report.html",
    "report_version_id": 1,
    "city": "",
    "model_aim": "",
    "model_name": "",
    "zakazchik": "",
    "high_level_department": "",
    "ds_name": "",
    "target_descr": "",
    "non_target_descr": "",
}


class ReportDecoWhitebox(ReportDeco):
    """
    Special report wrapper for :class:`~lightautoml.automl.presets.whitebox_presets.WhiteBoxPreset`.
    Usage case is the same as main
    :class:`~lightautoml.report.report_deco.ReportDeco` class.
    It generates same report as :class:`~lightautoml.report.report_deco.ReportDeco` ,
    but with additional whitebox report part.

    Difference:

        - report_automl.predict gets additional report argument.
          It stands for updating whitebox report part.
          Calling ``report_automl.predict(test_data, report=True)``
          will update test part of whitebox report.
          Calling ``report_automl.predict(test_data, report=False)``
          will extend general report with.
          New data and keeps whitebox part as is (much more faster).
        - :class:`~lightautoml.automl.presets.whitebox_presets.WhiteBoxPreset`
          should be created with parameter ``general_params={"report": True}``
          to get white box report part.
          If ``general_params`` set to ``{"report": False}``,
          only standard ReportDeco part will be created (much faster).

    """

    @property
    def model(self):
        """Get unwrapped WhiteBox.

        Returns:
            model.

        """
        # this is made to remove heavy whitebox inner report deco
        model = copy(self._model)
        try:
            model_wo_report = model.whitebox.model
        except AttributeError:
            return self._model

        pipe = copy(self._model.levels[0][0])
        ml_algo = copy(pipe.ml_algos[0])

        ml_algo.models = [model_wo_report]
        pipe.ml_algos = [ml_algo]

        model.levels = [[pipe]]

        return model

    @property
    def content(self):
        return self._model.whitebox._ReportDeco__stat

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.wb_report_params = copy(_default_wb_report_params)

        # self.wb_report_params = wb_report_params
        self.wb_report_params["output_path"] = self.output_path
        self._whitebox_section_path = "whitebox_section.html"
        self.sections_order.append("whitebox")

    def fit_predict(self, *args, **kwargs):
        """Wrapped :meth:`AutoML.fit_predict` method.

        Valid args, kwargs are the same as wrapped automl.

        Args:
            *args: Arguments.
            **kwargs: Additional parameters.

        Returns:
            OOF predictions.

        """
        predict_proba = super().fit_predict(*args, **kwargs)

        if self._model.general_params["report"]:
            self._generate_whitebox_section()
        else:
            logger.info2("Whitebox part is not created. Fit WhiteBox with general_params['report'] = True")

        self.generate_report()
        return predict_proba

    def predict(self, *args, **kwargs):
        """Wrapped :meth:`AutoML.predict` method.

        Valid args, kwargs are the same as wrapped automl.

        Args:
            *args: Arguments.
            **kwargs: Additional parameters.

        Returns:
            Predictions.

        """
        if len(args) >= 2:
            args = (args[0],)

        kwargs["report"] = self._model.general_params["report"]

        predict_proba = super().predict(*args, **kwargs)

        if self._model.general_params["report"]:
            self._generate_whitebox_section()
        else:
            logger.info2("Whitebox part is not created. Fit WhiteBox with general_params['report'] = True")

        self.generate_report()
        return predict_proba

    def _generate_whitebox_section(self):
        self._model.whitebox.generate_report(self.wb_report_params)
        content = self.content.copy()

        if self._n_test_sample >= 2:
            content["n_test_sample"] = self._n_test_sample
        content["model_coef"] = pd.DataFrame(content["model_coef"], columns=["Feature name", "Coefficient"]).to_html(
            index=False
        )
        content["p_vals"] = pd.DataFrame(content["p_vals"], columns=["Feature name", "P-value"]).to_html(index=False)
        content["p_vals_test"] = pd.DataFrame(content["p_vals_test"], columns=["Feature name", "P-value"]).to_html(
            index=False
        )
        content["train_vif"] = pd.DataFrame(content["train_vif"], columns=["Feature name", "VIF value"]).to_html(
            index=False
        )
        content["psi_total"] = pd.DataFrame(content["psi_total"], columns=["Feature name", "PSI value"]).to_html(
            index=False
        )
        content["psi_zeros"] = pd.DataFrame(content["psi_zeros"], columns=["Feature name", "PSI value"]).to_html(
            index=False
        )
        content["psi_ones"] = pd.DataFrame(content["psi_ones"], columns=["Feature name", "PSI value"]).to_html(
            index=False
        )

        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        self._sections["whitebox"] = env.get_template(self._whitebox_section_path).render(content)


def plot_data_hist(data, title="title", bins=100, path=None):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, axs = plt.subplots(figsize=(16, 10))
    sns.distplot(data, bins=bins, color="m", ax=axs)
    axs.set_title(title)
    fig.savefig(path, bbox_inches="tight")
    plt.close()


class ReportDecoNLP(ReportDeco):
    """
    Special report wrapper for :class:`~lightautoml.automl.presets.text_presets.TabularNLPAutoML`.
    Usage case is the same as main
    :class:`~lightautoml.report.report_deco.ReportDeco` class.
    It generates same report as :class:`~lightautoml.report.report_deco.ReportDeco` ,
    but with additional NLP report part.

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._nlp_section_path = "nlp_section.html"
        self._nlp_subsection_path = "nlp_subsection.html"
        self._nlp_subsections = []
        self.sections_order.append("nlp")

    def __call__(self, model):
        self._model = model

        # add informataion to report
        self._model_name = model.__class__.__name__
        self._model_parameters = json2html.convert(extract_params(model))
        self._model_summary = None

        self._sections = {}
        self._sections["intro"] = "<p>This report was generated automatically.</p>"
        self._model_results = []
        self._n_test_sample = 0

        self._generate_model_section()
        self.generate_report()
        return self

    def fit_predict(self, *args, **kwargs):
        """Wrapped :meth:`TabularNLPAutoML.fit_predict` method.

        Valid args, kwargs are the same as wrapped automl.

        Args:
            *args: Arguments.
            **kwargs: Additional parameters.

        Returns:
            OOF predictions.

        """
        preds = super().fit_predict(*args, **kwargs)

        train_data = kwargs["train_data"] if "train_data" in kwargs else args[0]
        roles = kwargs["roles"] if "roles" in kwargs else args[1]

        self._text_fields = self._get_text_fields(roles)
        train_data[self._text_fields] = train_data[self._text_fields].fillna("")
        for text_field in self._text_fields:
            content = {}
            content["title"] = "Text field: " + text_field
            content["char_len_hist"] = text_field + "_char_len_hist.png"
            plot_data_hist(
                data=train_data[text_field].apply(len).values,
                path=os.path.join(self.output_path, content["char_len_hist"]),
                title="Length in char",
            )
            content["tokens_len_hist"] = text_field + "_tokens_len_hist.png"
            plot_data_hist(
                data=train_data[text_field].str.split(" ").apply(len).values,
                path=os.path.join(self.output_path, content["tokens_len_hist"]),
                title="Length in tokens",
            )
            self._generate_nlp_subsection(content)
        # Concatenated text fields
        if len(self._text_fields) >= 2:
            all_fields = train_data[self._text_fields].agg(" ".join, axis=1)
            content = {}
            content["title"] = "Concatenated text fields"
            content["char_len_hist"] = "concat_char_len_hist.png"
            plot_data_hist(
                data=all_fields.apply(len).values,
                path=os.path.join(self.output_path, content["char_len_hist"]),
                title="Length in char",
            )
            content["tokens_len_hist"] = "concat_tokens_len_hist.png"
            plot_data_hist(
                data=all_fields.str.split(" ").apply(len).values,
                path=os.path.join(self.output_path, content["tokens_len_hist"]),
                title="Length in tokens",
            )
            self._generate_nlp_subsection(content)

        self._generate_nlp_section()
        self.generate_report()
        return preds

    def _generate_nlp_subsection(self, content):
        # content has the following fields:
        # title:            subsection title
        # char_len_hist:    path to histogram of text length (number of chars)
        # tokens_len_hist:  path to histogram of text length (number of tokens)
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        nlp_subsection = env.get_template(self._nlp_subsection_path).render(content)
        self._nlp_subsections.append(nlp_subsection)

    def _generate_nlp_section(self):
        if self._model_results:
            env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
            nlp_section = env.get_template(self._nlp_section_path).render(nlp_subsections=self._nlp_subsections)
            self._sections["nlp"] = nlp_section

    @staticmethod
    def _get_text_fields(roles: dict) -> list:
        """Returns all text fields, mentioned in roles.

        Args:
            roles: Roles.

        Returns:
            List of text fields.

        """
        text_roles = roles.get("text", [])
        for role_type, role_name in roles.items():
            if isinstance(role_type, laml_roles.TextRole):
                if isinstance(role_name, str):
                    text_roles.append(role_name)
                elif isinstance(role_name, list):
                    text_roles.extend(role_name)
        return text_roles


def get_uplift_data(test_target, uplift_pred, test_treatment, mode):
    try:
        perfect = uplift_metrics.perfect_uplift_curve(test_target, test_treatment)
        xs_perfect, ys_perfect = uplift_metrics.calculate_graphic_uplift_curve(
            test_target, perfect, test_treatment, mode
        )
    except NotImplementedError as e:  # noqa: F841
        xs_perfect, ys_perfect = None, None

    xs, ys = uplift_metrics.calculate_graphic_uplift_curve(test_target, uplift_pred, test_treatment, mode)
    normed = type_of_target(test_target) == "binary"
    uplift_auc = uplift_metrics.calculate_uplift_auc(test_target, uplift_pred, test_treatment, mode, normed=normed)
    return xs, ys, xs_perfect, ys_perfect, uplift_auc


def plot_uplift_curve(test_target, uplift_pred, test_treatment, path):
    sns.set(style="whitegrid", font_scale=1.5)
    # plt.figure(figsize=(10, 10));
    fig, axs = plt.subplots(3, 1, figsize=(10, 30))
    # qini
    xs, ys, xs_perfect, ys_perfect, uplift_auc = get_uplift_data(test_target, uplift_pred, test_treatment, mode="qini")
    axs[0].plot(xs, ys, color="blue", lw=2, label="qini mode")
    if xs_perfect is not None and ys_perfect is not None:
        axs[0].plot(xs_perfect, ys_perfect, color="black", lw=1, label="perfect uplift")
    axs[0].plot(
        (0, xs[-1]),
        (0, ys[-1]),
        color="black",
        lw=1,
        linestyle="--",
        label="random model",
    )
    axs[0].set_title("Uplift qini, AUC={:.3f}".format(uplift_auc))
    axs[0].legend(loc="lower right")
    # cum_gain
    xs, ys, xs_perfect, ys_perfect, uplift_auc = get_uplift_data(
        test_target, uplift_pred, test_treatment, mode="cum_gain"
    )
    axs[1].plot(xs, ys, color="red", lw=2, label="cum_gain model")
    if xs_perfect is not None and ys_perfect is not None:
        axs[1].plot(xs_perfect, ys_perfect, color="black", lw=1, label="perfect uplift")
    axs[1].plot(
        (0, xs[-1]),
        (0, ys[-1]),
        color="black",
        lw=1,
        linestyle="--",
        label="random model",
    )
    axs[1].set_title("Uplift cum_gain, AUC={:.3f}".format(uplift_auc))
    axs[1].legend(loc="lower right")
    # adj_qini
    xs, ys, xs_perfect, ys_perfect, uplift_auc = get_uplift_data(
        test_target, uplift_pred, test_treatment, mode="adj_qini"
    )
    axs[2].plot(xs, ys, color="green", lw=2, label="adj_qini mode")
    if xs_perfect is not None and ys_perfect is not None:
        axs[2].plot(xs_perfect, ys_perfect, color="black", lw=1, label="perfect uplift")
    axs[2].plot(
        (0, xs[-1]),
        (0, ys[-1]),
        color="black",
        lw=1,
        linestyle="--",
        label="random model",
    )
    axs[2].set_title("Uplift adj_qini, AUC={:.3f}".format(uplift_auc))
    axs[2].legend(loc="lower right")

    plt.savefig(path, bbox_inches="tight")
    plt.close()


class ReportDecoUplift(ReportDeco):
    _available_metalearners = (TLearner, XLearner)

    @property
    def reader(self):
        if self._is_xlearner:
            return self._model.learners["outcome"]["treatment"].reader  # effect
        else:
            return self._model.treatment_learner.reader

    @property
    def task(self):
        if self._is_xlearner:
            return "reg"
        else:
            return self.reader.task._name

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._uplift_section_path = "uplift_section.html"
        self._uplift_subsection_path = "uplift_subsection.html"
        self.sections_order.append("uplift")
        self._uplift_results = []

    def __call__(self, model):
        self._model = model
        self._is_xlearner = isinstance(model, XLearner)

        # add informataion to report
        self._model_name = model.__class__.__name__
        self._model_parameters = json2html.convert(extract_params(model))
        self._model_summary = None

        self._sections = {}
        self._sections["intro"] = "<p>This report was generated automatically.</p>"
        self._model_results = []
        self._n_test_sample = 0

        self._generate_model_section()
        self.generate_report()
        return self

    def fit(self, *args, **kwargs):
        """Wrapped automl.fit_predict method.

        Valid args, kwargs are the same as wrapped automl.

        Args:
            *args: arguments.
            **kwargs: additional parameters.

        """
        train_data = kwargs["train_data"] if "train_data" in kwargs else args[0]
        input_roles = kwargs["roles"] if "roles" in kwargs else args[1]
        self._target = input_roles["target"]
        self._treatment_col = input_roles["treatment"]
        if self._is_xlearner:
            self._fit_xlearner(train_data, input_roles)
        else:
            self._fit_tlearner(train_data, input_roles)
        self._model._is_fitted = True
        self._generate_model_section()
        self._train_data_overview = self._data_general_info(train_data, "train")
        self._describe_roles(train_data)
        self._describe_dropped_features(train_data)
        self._generate_train_set_section()
        self.generate_report()

    def predict(self, test_data):
        """Wrapped tlearner.predict method.

        Valid args, kwargs are the same as wrapped automl.

        Args:
            test_data: Dataset to perform inference.

        Returns:
            predictions.

        """
        self._n_test_sample += 1

        # get predictions
        test_target = test_data[self._target].values
        test_treatment = test_data[self._treatment_col].values
        # test_data = test_data.drop([self._target, self._treatment_col], axis=1)

        uplift, treatment_preds, control_preds = self._model.predict(test_data)

        if self._n_test_sample >= 2:
            treatment_title = "Treatment test {}".format(self._n_test_sample)
            control_title = "Control test {}".format(self._n_test_sample)
        else:
            treatment_title = "Treatment test"
            control_title = "Control test"

        # treatment data
        data = pd.DataFrame({"y_true": test_target[test_treatment == 1]})
        data["y_pred"] = treatment_preds[test_treatment == 1]
        data.sort_values("y_pred", ascending=False, inplace=True)
        data["bin"] = (np.arange(data.shape[0]) / data.shape[0] * self.n_bins).astype(int)
        data = data[~data["y_pred"].isnull()]
        self._generate_test_subsection(data, "treatment", treatment_title)
        self._generate_inference_section()

        # control data
        data = pd.DataFrame({"y_true": test_target[test_treatment == 0]})
        data["y_pred"] = control_preds[test_treatment == 0]
        data.sort_values("y_pred", ascending=False, inplace=True)
        data["bin"] = (np.arange(data.shape[0]) / data.shape[0] * self.n_bins).astype(int)
        data = data[~data["y_pred"].isnull()]
        self._generate_test_subsection(data, "control", control_title)
        self._generate_inference_section()

        # update model section
        self._generate_model_section()

        # uplift section
        self._uplift_content = {}
        if self._n_test_sample >= 2:
            self._uplift_content["title"] = "Test sample {}".format(self._n_test_sample)
            self._uplift_content["uplift_curve"] = "uplift_curve_{}.png".format(self._n_test_sample)
            self._uplift_content["uplift_distribution"] = "uplift_distribution_{}.png".format(self._n_test_sample)
        else:
            self._uplift_content["title"] = "Test sample"
            self._uplift_content["uplift_curve"] = "uplift_curve.png"
            self._uplift_content["uplift_distribution"] = "uplift_distribution.png"
        plot_uplift_curve(
            test_target,
            uplift,
            test_treatment,
            path=os.path.join(self.output_path, self._uplift_content["uplift_curve"]),
        )
        self._uplift_distribution(
            test_target,
            uplift,
            test_treatment,
            path=os.path.join(self.output_path, self._uplift_content["uplift_distribution"]),
        )

        self._uplift_content["test_data_overview"] = self._data_general_info(test_data, "test")

        self._generate_uplift_subsection()
        self._generate_uplift_section()

        self.generate_report()
        return uplift, treatment_preds, control_preds

    def _uplift_distribution(self, test_target, uplift, test_treatment, path):
        data = pd.DataFrame({"y_true": test_target, "y_pred": uplift, "treatment": test_treatment})
        data.sort_values("y_pred", ascending=True, inplace=True)
        data["bin"] = (np.arange(data.shape[0]) / data.shape[0] * self.n_bins).astype(int)
        # 'Uplift fact'
        mean_target_treatment = (
            data[data["treatment"].values == 1].groupby("bin").agg({"y_true": ["mean"]}).values[:, 0]
        )
        mean_target_control = data[data["treatment"].values == 0].groupby("bin").agg({"y_true": ["mean"]}).values[:, 0]
        uplift_fact = mean_target_treatment - mean_target_control
        bins_table = data.groupby("bin").agg({"y_true": [len], "y_pred": ["min", "mean", "max"]}).reset_index()
        bins_table.columns = [
            "Bin number",
            "Amount of objects",
            "Min uplift",
            "Mean uplift",
            "Max uplift",
        ]
        bins_table["Uplift fact"] = uplift_fact
        self._uplift_content["uplift_bins_table"] = bins_table.to_html(index=False)

        # uplift kde distribution
        sns.set(style="whitegrid", font_scale=1.5)
        fig, axs = plt.subplots(figsize=(16, 10))

        sns.kdeplot(data["y_pred"], fill=True, color="g", label="y_pred", ax=axs)
        axs.set_xlabel("Uplift value")
        axs.set_ylabel("Density")
        axs.set_title("Uplift distribution")

        fig.savefig(path, bbox_inches="tight")
        plt.close()

    def _fit_tlearner(self, train_data, roles):
        treatment_role, _ = _get_treatment_role(roles)
        new_roles = deepcopy(roles)
        new_roles.pop(treatment_role)
        self._model._timer._timeout = 1e10
        self._model._timer.start()
        # treatment
        treatment_train_data = train_data[train_data[self._treatment_col] == 1]
        treatment_target = treatment_train_data[self._target].values
        treatment_train_data.drop(self._treatment_col, axis=1, inplace=True)
        treatment_preds = self._model.treatment_learner.fit_predict(treatment_train_data, new_roles)
        # control
        control_train_data = train_data[train_data[self._treatment_col] == 0]
        control_target = control_train_data[self._target].values
        control_train_data.drop(self._treatment_col, axis=1, inplace=True)
        control_preds = self._model.control_learner.fit_predict(control_train_data, new_roles)

        self._generate_fit_section(treatment_preds, control_preds, treatment_target, control_target)

    def _fit_xlearner(self, train_data, roles):
        treatment_role, _ = _get_treatment_role(roles)
        new_roles = deepcopy(roles)
        new_roles.pop(treatment_role)

        self._model._timer._timeout = 1e10
        self._model._timer.start()
        self._model._fit_propensity_learner(train_data, roles)
        self._model._fit_outcome_learners(train_data, roles)

        # treatment
        treatment_train_data = train_data[train_data[self._treatment_col] == 1]
        treatment_train_data.drop(self._treatment_col, axis=1, inplace=True)
        outcome_pred = self._model.learners["outcome"]["control"].predict(treatment_train_data).data.ravel()
        treatment_train_data[self._target] = treatment_train_data[self._target] - outcome_pred
        treatment_target = treatment_train_data[self._target].values
        treatment_preds = self._model.learners["effect"]["treatment"].fit_predict(treatment_train_data, new_roles)

        # control
        control_train_data = train_data[train_data[self._treatment_col] == 0]
        control_train_data.drop(self._treatment_col, axis=1, inplace=True)
        outcome_pred = self._model.learners["outcome"]["treatment"].predict(control_train_data).data.ravel()
        control_train_data[self._target] = control_train_data[self._target] - outcome_pred
        control_train_data[self._target] *= -1
        control_target = control_train_data[self._target].values
        control_preds = self._model.learners["effect"]["control"].fit_predict(control_train_data, new_roles)

        self._generate_fit_section(treatment_preds, control_preds, treatment_target, control_target)

    def _generate_fit_section(self, treatment_preds, control_preds, treatment_target, control_target):
        self._generate_model_summary_table()
        # treatment model
        treatment_data = self._collect_data(treatment_preds, treatment_target)
        self._generate_training_subsection(treatment_data, "treatment", "Treatment train")
        self._generate_inference_section()

        control_data = self._collect_data(control_preds, control_target)
        self._generate_training_subsection(control_data, "control", "Control train")
        self._generate_inference_section()

    def _collect_data(self, preds, target):
        data = pd.DataFrame({"y_true": target})
        if self.task in "multiclass":
            if self.mapping is not None:
                data["y_true"] = np.array([self.mapping[y] for y in data["y_true"].values])
            data["y_pred"] = preds._data.argmax(axis=1)
        else:
            data["y_pred"] = preds._data[:, 0]
            data.sort_values("y_pred", ascending=False, inplace=True)
            data["bin"] = (np.arange(data.shape[0]) / data.shape[0] * self.n_bins).astype(int)
        # remove NaN in predictions:
        data = data[~data["y_pred"].isnull()]
        return data

    def _generate_model_summary_table(self):
        if self.task == "binary":
            evaluation_parameters = ["AUC-score", "Precision", "Recall", "F1-score"]
            self._model_summary = pd.DataFrame({"Evaluation parameter": evaluation_parameters})
        elif self.task == "reg":
            evaluation_parameters = [
                "Mean absolute error",
                "Median absolute error",
                "Mean squared error",
                "R^2 (coefficient of determination)",
                "Explained variance",
            ]
            self._model_summary = pd.DataFrame({"Evaluation parameter": evaluation_parameters})

    def _generate_training_subsection(self, data, prefix, title):
        self._inference_content = {}
        self._inference_content["title"] = title
        if self.task == "binary":
            # filling for html
            self._inference_content["roc_curve"] = prefix + "_roc_curve.png"
            self._inference_content["pr_curve"] = prefix + "_pr_curve.png"
            self._inference_content["pie_f1_metric"] = prefix + "_pie_f1_metric.png"
            self._inference_content["preds_distribution_by_bins"] = prefix + "_preds_distribution_by_bins.png"
            self._inference_content["distribution_of_logits"] = prefix + "_distribution_of_logits.png"
            # graphics and metrics
            _, self._F1_thresh = f1_score_w_co(data)
            self._model_summary[title] = self._binary_classification_details(data)
        elif self.task == "reg":
            # filling for html
            self._inference_content["target_distribution"] = prefix + "_target_distribution.png"
            self._inference_content["error_hist"] = prefix + "_error_hist.png"
            self._inference_content["scatter_plot"] = prefix + "_scatter_plot.png"
            # graphics and metrics
            self._model_summary[title] = self._regression_details(data)

    def _generate_test_subsection(self, data, prefix, title):
        self._inference_content = {}
        self._inference_content["title"] = title
        if self.task == "binary":
            # filling for html
            self._inference_content["roc_curve"] = prefix + "_roc_curve_{}.png".format(self._n_test_sample)
            self._inference_content["pr_curve"] = prefix + "_pr_curve_{}.png".format(self._n_test_sample)
            self._inference_content["pie_f1_metric"] = prefix + "_pie_f1_metric_{}.png".format(self._n_test_sample)
            self._inference_content["bins_preds"] = prefix + "_bins_preds_{}.png".format(self._n_test_sample)
            self._inference_content[
                "preds_distribution_by_bins"
            ] = prefix + "_preds_distribution_by_bins_{}.png".format(self._n_test_sample)
            self._inference_content["distribution_of_logits"] = prefix + "_distribution_of_logits_{}.png".format(
                self._n_test_sample
            )
            # graphics and metrics
            self._model_summary[title] = self._binary_classification_details(data)
        elif self.task == "reg":
            # filling for html
            self._inference_content["target_distribution"] = prefix + "_target_distribution_{}.png".format(
                self._n_test_sample
            )
            self._inference_content["error_hist"] = prefix + "_error_hist_{}.png".format(self._n_test_sample)
            self._inference_content["scatter_plot"] = prefix + "_scatter_plot_{}.png".format(self._n_test_sample)
            # graphics
            self._model_summary[title] = self._regression_details(data)

    def _data_general_info(self, data, stage="train"):
        general_info = pd.DataFrame(columns=["Parameter", "Value"])
        general_info.loc[0] = ("Number of records", data.shape[0])
        general_info.loc[1] = ("Share of treatment", np.mean(data[self._treatment_col]))
        general_info.loc[2] = ("Mean target", np.mean(data[self._target]))
        general_info.loc[3] = (
            "Mean target on treatment",
            np.mean(data[self._target][data[self._treatment_col] == 1]),
        )
        general_info.loc[4] = (
            "Mean target on control",
            np.mean(data[self._target][data[self._treatment_col] == 0]),
        )
        if stage == "train":
            general_info.loc[5] = ("Total number of features", data.shape[1])
            general_info.loc[6] = ("Used features", len(self.reader._used_features))
            dropped_list = [col for col in self.reader._dropped_features if col != self._target]
            general_info.loc[7] = ("Dropped features", len(dropped_list))
        return general_info.to_html(index=False, justify="left")

    def _describe_roles(self, train_data):

        # detect feature roles
        # roles = self._model.reader._roles
        roles = self.reader._roles
        numerical_features = [feat_name for feat_name in roles if roles[feat_name].name == "Numeric"]
        categorical_features = [feat_name for feat_name in roles if roles[feat_name].name == "Category"]
        datetime_features = [feat_name for feat_name in roles if roles[feat_name].name == "Datetime"]
        text_features = [feat_name for feat_name in roles if roles[feat_name].name == "Text"]

        # numerical roles
        numerical_features_df = []
        for feature_name in numerical_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(train_data[feature_name].isna().sum() / train_data.shape[0])
            values = train_data[feature_name].dropna().values
            item["min"] = np.min(values)
            item["quantile_25"] = np.quantile(values, 0.25)
            item["average"] = np.mean(values)
            item["median"] = np.median(values)
            item["quantile_75"] = np.quantile(values, 0.75)
            item["max"] = np.max(values)
            numerical_features_df.append(item)
        if numerical_features_df == []:
            self._numerical_features_table = None
        else:
            self._numerical_features_table = pd.DataFrame(numerical_features_df).to_html(
                index=False, float_format="{:.2f}".format, justify="left"
            )
        # categorical roles
        categorical_features_df = []
        for feature_name in categorical_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(train_data[feature_name].isna().sum() / train_data.shape[0])
            value_counts = train_data[feature_name].value_counts(normalize=True)
            values = value_counts.index.values
            counts = value_counts.values
            item["Number of unique values"] = len(counts)
            item["Most frequent value"] = values[0]
            item["Occurrence of most frequent"] = "{:.1f}%".format(100 * counts[0])
            item["Least frequent value"] = values[-1]
            item["Occurrence of least frequent"] = "{:.1f}%".format(100 * counts[-1])
            categorical_features_df.append(item)
        if categorical_features_df == []:
            self._categorical_features_table = None
        else:
            self._categorical_features_table = pd.DataFrame(categorical_features_df).to_html(
                index=False, justify="left"
            )
        # datetime roles
        datetime_features_df = []
        for feature_name in datetime_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(train_data[feature_name].isna().sum() / train_data.shape[0])
            values = train_data[feature_name].dropna().values
            item["min"] = np.min(values)
            item["max"] = np.max(values)
            item["base_date"] = self.reader._roles[feature_name].base_date
            datetime_features_df.append(item)
        if datetime_features_df == []:
            self._datetime_features_table = None
        else:
            self._datetime_features_table = pd.DataFrame(datetime_features_df).to_html(index=False, justify="left")
        # text roles
        text_features_df = []
        for feature_name in text_features:
            item = {"Feature name": feature_name}
            feature_length = train_data[feature_name].str.len()
            item["Amount of empty records"] = (feature_length == 0).sum(axis=0)
            item["Length of the shortest sentence"] = feature_length.min()
            item["Length of the longest sentence"] = feature_length.max()
            text_features_df.append(item)
        self._text_features_table = list2table(text_features_df)

    def _describe_dropped_features(self, train_data):
        self._max_nan_rate = self.reader.max_nan_rate
        self._max_constant_rate = self.reader.max_constant_rate
        self._features_dropped_list = self.reader._dropped_features
        # dropped features table
        dropped_list = [col for col in self._features_dropped_list if col != self._target]
        if dropped_list == []:
            self._dropped_features_table = None
        else:
            dropped_nan_ratio = train_data[dropped_list].isna().sum() / train_data.shape[0]
            dropped_most_occured = pd.Series(np.nan, index=dropped_list)
            for col in dropped_list:
                col_most_occured = train_data[col].value_counts(normalize=True).values
                if len(col_most_occured) > 0:
                    dropped_most_occured[col] = col_most_occured[0]
            dropped_features_table = pd.DataFrame(
                {"nan_rate": dropped_nan_ratio, "constant_rate": dropped_most_occured}
            )
            self._dropped_features_table = (
                dropped_features_table.reset_index()
                .rename(columns={"index": "Название переменной"})
                .to_html(index=False, justify="left")
            )

    def _generate_train_set_section(self):
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        train_set_section = env.get_template(self._train_set_section_path).render(
            train_data_overview=self._train_data_overview,
            numerical_features_table=self._numerical_features_table,
            categorical_features_table=self._categorical_features_table,
            datetime_features_table=self._datetime_features_table,
            target=self._target,
            max_nan_rate=self._max_nan_rate,
            max_constant_rate=self._max_constant_rate,
            dropped_features_table=self._dropped_features_table,
        )
        self._sections["train_set"] = train_set_section

    def _generate_uplift_subsection(self):
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        uplift_subsection = env.get_template(self._uplift_subsection_path).render(self._uplift_content)
        self._uplift_results.append(uplift_subsection)

    def _generate_uplift_section(self):
        if self._model_results:
            env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
            results_section = env.get_template(self._uplift_section_path).render(uplift_results=self._uplift_results)
            self._sections["uplift"] = results_section



"""Metrics and loss functions for LightGBM."""

import logging

from functools import partial
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import lightgbm as lgb
import numpy as np

from ..common_metric import _valid_str_multiclass_metric_names
from ..utils import infer_gib
from .base import Loss
from .base import fw_rmsle
from .lgb_custom import lgb_f1_loss_multiclass  # , F1Factory
from .lgb_custom import softmax_ax1


logger = logging.getLogger(__name__)


_lgb_binary_metrics_dict = {
    "auc": "auc",
    "logloss": "binary_logloss",
    "accuracy": "binary_error",
}

_lgb_reg_metrics_dict = {
    "mse": "mse",
    "mae": "mae",
    "r2": "mse",
    "rmsle": "mse",
    "mape": "mape",
    "quantile": "quantile",
    "huber": "huber",
    "fair": "fair",
}

_lgb_multiclass_metrics_dict = {
    "auc": _valid_str_multiclass_metric_names["auc"],
    "auc_mu": _valid_str_multiclass_metric_names["auc_mu"],
    "crossentropy": "multi_logloss",
    "accuracy": "multi_error",
    "f1_macro": _valid_str_multiclass_metric_names["f1_macro"],
    "f1_micro": _valid_str_multiclass_metric_names["f1_micro"],
    "f1_weighted": _valid_str_multiclass_metric_names["f1_weighted"],
}


_lgb_metrics_dict = {
    "binary": _lgb_binary_metrics_dict,
    "reg": _lgb_reg_metrics_dict,
    "multiclass": _lgb_multiclass_metrics_dict,
}

_lgb_loss_mapping = {
    "logloss": ("binary", None, None),
    "mse": ("regression", None, None),
    "mae": ("l1", None, None),
    "mape": ("mape", None, None),
    "crossentropy": ("multiclass", None, None),
    "rmsle": ("mse", fw_rmsle, np.expm1),
    "quantile": ("quantile", None, None),
    "huber": ("huber", None, None),
    "fair": ("fair", None, None),
    "f1": (lgb_f1_loss_multiclass, None, softmax_ax1),
}

_lgb_loss_params_mapping = {
    "quantile": {"q": "alpha"},
    "huber": {"a": "alpha"},
    "fair_c": {"c": "fair_c"},
}

_lgb_force_metric = {
    "rmsle": ("mse", None, None),
}


class LGBFunc:
    """Wrapper of metric function for LightGBM."""

    def __init__(self, metric_func, greater_is_better, bw_func):
        self.metric_func = metric_func
        self.greater_is_better = greater_is_better
        self.bw_func = bw_func

    def __call__(self, pred: np.ndarray, dtrain: lgb.Dataset) -> Tuple[str, float, bool]:
        """Calculate metric."""
        label = dtrain.get_label()

        weights = dtrain.get_weight()

        if label.shape[0] != pred.shape[0]:
            pred = pred.reshape((label.shape[0], -1), order="F")
            label = label.astype(np.int32)

        label = self.bw_func(label)
        pred = self.bw_func(pred)

        # for weighted case
        try:
            val = self.metric_func(label, pred, sample_weight=weights)
        except TypeError:
            val = self.metric_func(label, pred)

        # TODO: what if grouped case

        return "Opt metric", val, self.greater_is_better


class LGBLoss(Loss):
    """Loss used for LightGBM.

    Args:
        loss: Objective to optimize.
        loss_params: additional loss parameters.
            Format like in :mod:`lightautoml.tasks.custom_metrics`.
        fw_func: forward transformation.
            Used for transformation of target and item weights.
        bw_func: backward transformation.
            Used for predict values transformation.

    Note:
        Loss can be one of the types:

            - Str: one of default losses
                ('auc', 'mse', 'mae', 'logloss', 'accuray', 'r2',
                'rmsle', 'mape', 'quantile', 'huber', 'fair')
                or another lightgbm objective.
            - Callable: custom lightgbm style objective.

    """

    def __init__(
        self,
        loss: Union[str, Callable],
        loss_params: Optional[Dict] = None,
        fw_func: Optional[Callable] = None,
        bw_func: Optional[Callable] = None,
    ):
        self.metric = None
        self.fobj = None
        self.fobj_name = None

        if loss in _lgb_loss_mapping:
            fobj, fw_func, bw_func = _lgb_loss_mapping[loss]
            if type(fobj) is str:
                self.fobj_name = fobj
            else:
                self.fobj = fobj
            # map param name for known objectives
            if self.fobj_name in _lgb_loss_params_mapping:
                param_mapping = _lgb_loss_params_mapping[self.fobj_name]
                loss_params = {param_mapping[x]: loss_params[x] for x in loss_params}

        else:
            # set lgb style objective
            if type(loss) is str:
                self.fobj_name = loss
            else:
                self.fobj = loss

        # set forward and backward transformations
        if fw_func is not None:
            self._fw_func = fw_func
        if bw_func is not None:
            self._bw_func = bw_func

        self.fobj_params = {}
        if loss_params is not None:
            self.fobj_params = loss_params

        self.metric = None

    def metric_wrapper(
        self,
        metric_func: Callable,
        greater_is_better: Optional[bool],
        metric_params: Optional[Dict] = None,
    ) -> Callable:
        """Customize metric.

        Args:
            metric_func: Callable metric.
            greater_is_better: Whether or not higher value is better.
            metric_params: Additional metric parameters.

        Returns:
            Callable metric, that returns ('Opt metric', value, greater_is_better).

        """
        if greater_is_better is None:
            greater_is_better = infer_gib(metric_func)

        if metric_params is not None:
            metric_func = partial(metric_func, **metric_params)

        return LGBFunc(metric_func, greater_is_better, self._bw_func)

    def set_callback_metric(
        self,
        metric: Union[str, Callable],
        greater_is_better: Optional[bool] = None,
        metric_params: Optional[Dict] = None,
        task_name: Optional[str] = None,
    ):
        """Callback metric setter.

        Args:
            metric: Callback metric.
            greater_is_better: Whether or not higher value is better.
            metric_params: Additional metric parameters.
            task_name: Name of task.

        Note:
            Value of ``task_name`` should be one of following options:

            - `'binary'`
            - `'reg'`
            - `'multiclass'`

        """
        # force metric if special loss
        # what about task_name? in this case?
        if self.fobj_name in _lgb_force_metric:
            metric, greater_is_better, metric_params = _lgb_force_metric[self.fobj_name]
            logger.info2(
                "For lgbm {0} callback metric switched to {1}".format(self.fobj_name, metric),
                UserWarning,
            )

        self.metric_params = {}

        # set lgb style metric
        self.metric = metric
        if type(metric) is str:

            if metric_params is not None:
                self.metric_params = metric_params

            _metric_dict = _lgb_metrics_dict[task_name]
            _metric = _metric_dict.get(metric)
            if type(_metric) is str:
                self.metric_name = _metric
                self.feval = None
            else:
                self.metric_name = None
                # _metric = CustomWrapper(_metric)
                self.feval = self.metric_wrapper(_metric, greater_is_better, {})

        else:
            self.metric_name = None
            # metric = CustomWrapper(metric)
            self.feval = self.metric_wrapper(metric, greater_is_better, self.metric_params)



"""Timer."""

import logging

from time import time
from typing import List
from typing import Optional
from typing import Union

import numpy as np

from .logging import DuplicateFilter


logger = logging.getLogger(__name__)
logger.addFilter(DuplicateFilter())


class Timer:
    """Timer to limit the duration tasks."""

    _timeout = 1e10
    _overhead = 0
    _mode = 1

    def __init__(self):
        self.start_time = None

    @property
    def time_left(self) -> float:
        if self.time_spent is not None:
            return self.timeout - self.time_spent
        return None

    @property
    def time_spent(self) -> float:
        if self.start_time is not None:
            return time() - self.start_time
        return None

    @property
    def perc_left(self) -> float:
        if self.time_left is not None:
            return self.time_left / self.timeout
        return None

    @property
    def perc_spent(self) -> float:
        if self.time_spent is not None:
            return self.time_spent / self.timeout
        return None

    @property
    def timeout(self) -> float:
        return self._timeout

    def time_limit_exceeded(self) -> bool:
        if self._mode == 0:
            return False

        if self._mode == 1:
            return self.time_left < 0

        if self._mode == 2:
            if self.time_left:
                return (self.time_left - self._overhead) < 0
            return None

    def start(self):
        self.start_time = time()
        return self


class PipelineTimer(Timer):
    """Timer is used to control time over full automl run.

    It decides how much time spend to each algo


    Args:
        timeout: Maximum amount of time that AutoML can run.
        overhead: (0, 1) - Rate of time that will be used to early stop.
            Ex. if set to `0.1` and timing mode is set to 2,
            timer will finish tasks after `0.9` of all time spent.
        mode: Timing mode. Can be 0, 1 or 2.
            Keep in mind - all time limitations will
            turn on after at least single model/single fold will be computed.
        tuning_rate: Approximate fraction of all time will be used for tuning.

    Note:
        Modes explanation:

            - 0 - timer is used to estimate runtime,
                but if something goes out of time,
                keep it run (Real life mode).
            - 1 - timer is used to terminate tasks,
                but do it after real timeout (Trade off mode).
            - 2 - timer is used to terminate tasks
                with the goal to be exactly
                in time (Benchmarking/competitions mode).

    """

    def __init__(
        self,
        timeout: Optional[float] = None,
        overhead: float = 0.1,
        mode: int = 1,
        tuning_rate: float = 0.7,
    ):
        if timeout is not None:
            self._timeout = timeout
        self._task_scores = 0
        self._rate_overhead = overhead
        self._overhead = overhead * self.timeout
        self.run_info = {}
        self.run_scores = {}
        self._mode = mode
        self.tuning_rate = tuning_rate
        self.child_out_of_time = False

    def add_task(self, score: float = 1.0):
        self._task_scores += score

    def close_task(self, score: float = 1.0):
        self._task_scores -= score

    def get_time_for_next_task(self, score: float = 1.0):
        if round(self._task_scores, 3) == 0:
            return self.time_left

        return (self.time_left - self._overhead) * (score / self._task_scores)

    def get_task_timer(self, key: Optional[str] = None, score: float = 1.0) -> "TaskTimer":
        return TaskTimer(self, key, score, self._rate_overhead, self._mode, self.tuning_rate)


class TaskTimer(Timer):
    """Timer is used to control time over single ML task run.

    It decides how much time is ok to spend on tuner
    and if we have enough time to calc more folds.


    Args:
        pipe_timer: Global automl timer.
        key: String name that will be associated with this task.
        score: Time score for current task.
            For ex. if you want to give more
            of total time to task set it > 1.
        overhead: See overhead of :class:`~lightautoml.utils.timer.PipelineTimer`.
        mode: See mode for :class:`~lightautoml.utils.timer.PipelineTimer`.
        default_tuner_time_rate: If no timing history for the moment
            of estimating tuning time,
            timer will use this rate of `time_left`.

    """

    @property
    def in_progress(self) -> bool:
        """Check if the task is running."""
        return self.start_time is not None

    def __init__(
        self,
        pipe_timer: PipelineTimer,
        key: Optional[str] = None,
        score: float = 1.0,
        overhead: Optional[float] = 1,
        mode: int = 1,
        default_tuner_time_rate: float = 0.7,
    ):
        self.score = score
        pipe_timer.add_task(self.score)
        self.pipe_timer = pipe_timer
        self.start_time = None
        self.key = key
        self._rate_overhead = overhead
        self._mode = mode
        self.default_tuner_rate = default_tuner_time_rate

    def start(self):
        """Starts counting down.

        Returns:
            self.

        """
        if self.in_progress:
            return self

        self.start_time = time()
        self._timeout = self.pipe_timer.get_time_for_next_task(self.score)
        self._overhead = self._rate_overhead * self.time_left
        self.pipe_timer.close_task(self.score)

        return self

    def set_control_point(self):
        """Set control point.

        Updates the countdown and time left parameters.

        """
        self._timeout = self.timeout - self.time_spent
        self.start_time = time()

    def write_run_info(self):
        """Collect timer history."""
        if self.key in self.pipe_timer.run_info:
            self.pipe_timer.run_info[self.key].append(self.time_spent)
            self.pipe_timer.run_scores[self.key].append(self.score)
        else:
            self.pipe_timer.run_info[self.key] = [self.time_spent]
            self.pipe_timer.run_scores[self.key] = [self.score]

    def get_run_results(self) -> Union[None, np.ndarray]:
        """Get timer history.

        Returns:
            ``None`` if there is no history, or array with history of runs.

        """
        if self.key in self.pipe_timer.run_info:
            return self.pipe_timer.run_info[self.key]

    def get_run_scores(self) -> Union[None, np.ndarray]:
        """Get timer scores.

        Returns:
            ``None`` if there is no scores, or array with scores of runs.

        """
        if self.key in self.pipe_timer.run_scores:
            return self.pipe_timer.run_scores[self.key]

    def estimate_folds_time(self, n_folds: int = 1) -> Optional[float]:
        """Estimate time for n_folds.

        Args:
            n_folds: Number of folds.

        Returns:
            Estimated time needed to run all `n_folds`.

        """
        run_results, run_scores = self.get_run_results(), self.get_run_scores()
        if run_results is None:
            if self._mode > 0:
                return None
            # case - at least one algo runs before and timer mode set to 0 (conservative mode)
            total_run_info, total_run_scores = [], []
            for k in self.pipe_timer.run_info:
                total_run_info.extend(self.pipe_timer.run_info[k])
                total_run_scores.extend(self.pipe_timer.run_scores[k])

            if len(total_run_info) == 0:
                return None

            single_run_est = np.array(total_run_info).sum() / np.array(total_run_scores).sum() * self.score
            return single_run_est * n_folds

        # case - algo runs at least ones
        if self._mode > 0:
            single_run_est = np.max(np.array(run_results) / np.array(run_scores))
        else:
            single_run_est = np.mean(np.array(run_results) / np.array(run_scores))

        single_run_est = single_run_est * self.score
        return single_run_est * n_folds

    def estimate_tuner_time(self, n_folds: int = 1) -> float:
        """Estimates time that is ok to spend on tuner.

        Args:
            n_folds: Number of folds.

        Returns:
            How much time timer will be able spend on tuner.

        """
        folds_est = self.estimate_folds_time(n_folds)
        if folds_est is None:
            if self.time_left:
                return self.default_tuner_rate * self.time_left
            else:
                return None
        return self.time_left - folds_est

    def time_limit_exceeded(self) -> bool:
        """Estimate time limit and send results to parent timer.

        Returns:
            ``True`` if time limit exceeded.

        """
        out_of_time = super().time_limit_exceeded()
        if out_of_time:
            self.pipe_timer.child_out_of_time = True
        return out_of_time

    def __copy__(self):

        proxy_timer = PipelineTimer().start()
        logger.warning(
            "Copying TaskTimer may affect the parent PipelineTimer, so copy will create new unlimited TaskTimer"
        )

        return proxy_timer.get_task_timer(self.key)

    def __deepcopy__(self, *args, **kwargs):

        return self.__copy__()

    def split_timer(self, n_parts: int) -> List["TaskTimer"]:
        """Split the timer into equal-sized tasks.

        Args:
             n_parts: Number of tasks.

        Returns:
            Timers for each tasks.

        """
        new_tasks_score = self.score / n_parts
        timers = [self.pipe_timer.get_task_timer(self.key, new_tasks_score) for _ in range(n_parts)]
        self.pipe_timer.close_task(self.score)

        return timers



"""Module for Ghost Batch Norm and variations.

Ghost Batch Norm: https://arxiv.org/pdf/1705.08741.pdf

"""

from math import ceil
from typing import Union

import torch
from torch import Tensor
from torch import nn


class GhostNorm(nn.Module):
    """Ghost Normalization.

    https://arxiv.org/pdf/1705.08741.pdf

    Args:
        inner_norm : torch.nn.Module (initialiezd)
            examples: `nn.BatchNorm1d`, `nn.LayerNorm`
        virtual_batch_size : int
        device : string or torch.device, optional
            default is "cpu"
    """

    def __init__(
        self,
        inner_norm: nn.Module,
        virtual_batch_size: int,
        device: Union[str, torch.device] = "cpu",
    ):
        super().__init__()
        self.virtual_batch_size = virtual_batch_size
        self.inner_norm = inner_norm
        self.to(device)

    def forward(self, x: Tensor) -> Tensor:
        """Transform the input tensor.

        Args:
            x : torch.Tensor

        Returns:
            torch.Tensor

        """
        chunk_size = int(ceil(x.shape[0] / self.virtual_batch_size))
        chunk_norm = [self.inner_norm(chunk) for chunk in x.chunk(chunk_size, dim=0)]
        return torch.cat(chunk_norm, dim=0)


class GhostBatchNorm(GhostNorm):
    """Ghost Normalization, using BatchNorm1d as inner normalization.

    https://arxiv.org/pdf/1705.08741.pdf

    Args:
        num_features : int
        virtual_batch_size : int, optional
            default is 64
        momentum : float, optional
            default is 0.1
        device : string or torch.device, optional
            default is "cpu"
    """

    def __init__(
        self,
        num_features: int,
        virtual_batch_size: int = 64,
        momentum: float = 0.1,
        device: Union[str, torch.device] = "cpu",
    ):
        super().__init__(
            inner_norm=nn.BatchNorm1d(num_features, momentum=momentum),
            virtual_batch_size=virtual_batch_size,
        )



"""PyTorch modules for the AutoInt model."""
# Paper: https://arxiv.org/pdf/1810.11921v2.pdf
# Official implementation: https://github.com/DeepGraphLearning/RecommenderSystems

from collections import namedtuple
from typing import Optional, Type, Union
import torch
from torch import nn, Tensor
from torch.nn import functional as F


EmbeddingInfo = namedtuple("EmbeddingInfo", ["num_fields", "output_size"])
UniformEmbeddingInfo = namedtuple("EmbeddingInfo", ["num_fields", "embedding_size", "output_size"])


class LeakyGate(nn.Module):
    """LeakyGate from https://github.com/jrfiedler/xynn.

    This performs an element-wise linear transformation followed by a chosen
    activation; the default activation is nn.LeakyReLU. Fields may be
    represented by individual values or vectors of values (i.e., embedded).

    Input needs to be shaped like (num_rows, num_fields) or
    (num_rows, num_fields, embedding_size)

    Args:
        input_size: input_size.
        bias: if to use bias.
        activation: activation function.
        device: device.
    """

    def __init__(
        self,
        input_size: int,
        bias: bool = True,
        activation: Type[nn.Module] = nn.LeakyReLU,
        device: Union[str, torch.device] = "cpu",
    ):
        super().__init__()
        self.weight = nn.Parameter(torch.normal(mean=0, std=1.0, size=(1, input_size)))
        self.bias = nn.Parameter(torch.zeros(size=(1, input_size)), requires_grad=bias)
        self.activation = activation()
        self.to(device)

    def forward(self, X: Tensor) -> Tensor:
        """Transform the input tensor.

        Args:
            X : torch.Tensor

        Returns:
            torch.Tensor
        """
        out = X
        if len(X.shape) > 2:
            out = out.reshape((X.shape[0], -1))
        out = out * self.weight + self.bias
        if len(X.shape) > 2:
            out = out.reshape(X.shape)
        out = self.activation(out)
        return out


def _initialized_tensor(*sizes):
    weight = nn.Parameter(torch.Tensor(*sizes))
    nn.init.kaiming_uniform_(weight)
    return weight


class AttnInteractionLayer(nn.Module):
    """The attention interaction layer for the AutoInt model.

    Paper for the original AutoInt model: https://arxiv.org/pdf/1810.11921v2.pdf

    Args:
        field_input_size : int
            original embedding size for each field
        field_output_size : int, optional
            embedding size after transformation; default is 8
        num_heads : int, optional
            number of attention heads; default is 2
        activation : subclass of torch.nn.Module or None, optional
            applied to the W tensors; default is None
        use_residual : bool, optional
            default is True
        dropout : float, optional
            default is 0.1
        normalize : bool, optional
            default is True
        ghost_batch_size : int or None, optional
            only used if `use_bn` is True; size of batch in "ghost batch norm";
            if None, normal batch norm is used; default is None
        device : string or torch.device, optional
            default is "cpu"

    """

    def __init__(
        self,
        field_input_size: int,
        field_output_size: int = 8,
        num_heads: int = 2,
        activation: Optional[Type[nn.Module]] = None,
        use_residual: bool = True,
        dropout: float = 0.1,
        normalize: bool = True,
        ghost_batch_size: Optional[int] = None,
        device: Union[str, torch.device] = "cpu",
    ):
        super().__init__()

        self.use_residual = use_residual

        self.W_q = _initialized_tensor(field_input_size, field_output_size, num_heads)
        self.W_k = _initialized_tensor(field_input_size, field_output_size, num_heads)
        self.W_v = _initialized_tensor(field_input_size, field_output_size, num_heads)

        if use_residual:
            self.W_r = _initialized_tensor(field_input_size, field_output_size * num_heads)
        else:
            self.W_r = None

        if activation:
            self.w_act = activation()
        else:
            self.w_act = nn.Identity()

        if dropout > 0.0:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = nn.Identity()

        if normalize:
            self.layer_norm = nn.LayerNorm(field_output_size * num_heads)
        else:
            self.layer_norm = nn.Identity()

        self.to(device)

    def forward(self, x: Tensor) -> Tensor:
        """Transform the input tensor with attention interaction.

        Args:
            x : torch.Tensor
                3-d tensor; for example, embedded numeric and/or categorical values,
                or the output of a previous attention interaction layer

        Returns:
            torch.Tensor

        """
        # R : # rows
        # F, D : # fields
        # I : field embedding size in
        # O : field embedding size out
        # H : # heads
        num_rows, num_fields, _ = x.shape  # R, F, I

        # (R, F, I) * (I, O, H) -> (R, F, O, H)
        qrys = torch.tensordot(x, self.w_act(self.W_q), dims=([-1], [0]))
        keys = torch.tensordot(x, self.w_act(self.W_k), dims=([-1], [0]))
        vals = torch.tensordot(x, self.w_act(self.W_v), dims=([-1], [0]))
        if self.use_residual:
            rsdl = torch.tensordot(x, self.w_act(self.W_r), dims=([-1], [0]))

        product = torch.einsum("rdoh,rfoh->rdfh", qrys, keys)  # (R, F, F, H)

        alpha = F.softmax(product, dim=2)  # (R, F, F, H)
        alpha = self.dropout(alpha)

        # (R, F, F, H) * (R, F, O, H) -> (R, F, O, H)
        out = torch.einsum("rfdh,rfoh->rfoh", alpha, vals)
        out = out.reshape((num_rows, num_fields, -1))  # (R, F, O * H)
        if self.use_residual:
            out = out + rsdl  # (R, F, O * H)
        out = F.leaky_relu(out)
        out = self.layer_norm(out)

        return out


class AttnInteractionBlock(nn.Module):
    """A collection of AttnInteractionLayers, followed by an optional "leaky gate" and then a linear layer.

    This block is originally for the AutoInt model.

    Code from: https://github.com/jrfiedler/xynn

    Args:
        field_input_size : int
            original embedding size for each field
        field_output_size : int, optional
            embedding size after transformation; default is 8
        num_layers : int, optional
            number of attention layers; default is 3
        num_heads : int, optional
            number of attention heads per layer; default is 2
        activation : subclass of torch.nn.Module or None, optional
            applied to the W tensors; default is None
        use_residual : bool, optional
            default is True
        dropout : float, optional
            default is 0.0
        normalize : bool, optional
            default is True
        ghost_batch_size : int or None, optional
            only used if `use_bn` is True; size of batch in "ghost batch norm";
            if None, normal batch norm is used; default is None
        device : string or torch.device, optional
            default is "cpu"
    """

    def __init__(
        self,
        field_input_size: int,
        field_output_size: int = 8,
        num_layers: int = 3,
        num_heads: int = 2,
        activation: Optional[Type[nn.Module]] = None,
        use_residual: bool = True,
        dropout: float = 0.1,
        normalize: bool = True,
        ghost_batch_size: Optional[int] = None,
        device: Union[str, torch.device] = "cpu",
    ):
        super().__init__()

        layers = []
        for _ in range(num_layers):
            layers.append(
                AttnInteractionLayer(
                    field_input_size,
                    field_output_size,
                    num_heads,
                    activation,
                    use_residual,
                    dropout,
                    normalize,
                    ghost_batch_size,
                    device,
                )
            )
            field_input_size = field_output_size * num_heads

        self.layers = nn.Sequential(*layers)
        self.to(device)

    def forward(self, x: Tensor) -> Tensor:
        """Transform the input tensor.

        Args:
            x : torch.Tensor
                3-d tensor, usually embedded numeric and/or categorical values

        Returns:
            torch.Tensor
        """
        out = self.layers(x)
        return out



"""Modules with machine learning algorithms and hyperparameters tuning tools."""

__all__ = [
    "tuning",
    "base",
    "boost_lgbm",
    "boost_cb",
    "linear_sklearn",
    "dl_model",
    "torch_based",
    "whitebox",
    "utils",
]



"""Image utils."""

from PIL import Image


def pil_loader(path: str) -> Image:
    """Load image from paths.

    Args:
        path: Image path.

    Returns:
        Loaded PIL Image in rgb.

    """
    with open(path, "rb") as f:
        img = Image.open(f)
        return img.convert("RGB")



"""Weighted average transformer for sequence embeddings."""

import logging

from collections import Counter
from itertools import repeat
from typing import Any
from typing import Dict
from typing import Sequence

import numpy as np

from scipy.linalg import svd
from sklearn.base import TransformerMixin
from sklearn.feature_extraction import DictVectorizer
from tqdm import tqdm


logger = logging.getLogger(__name__)


class WeightedAverageTransformer(TransformerMixin):
    """Weighted average of word embeddings.

    Calculate sentence embedding as weighted average of word embeddings.

    Args:
        embedding_model: word2vec, fasstext, etc.
            Should have dict interface {<word>: <embedding>}.
        embed_size: Size of embedding.
        weight_type: 'idf' for idf weights, 'sif' for
            smoothed inverse frequency weights, '1' for all weights are equal.
        use_svd: Subtract projection onto first singular vector.
        alpha: Param for sif weights.
        verbose: Add prints.
        **kwargs: Unused arguments.

    """

    name = "WAT"

    def __init__(
        self,
        embedding_model: Dict,
        embed_size: int,
        weight_type: str = "idf",
        use_svd: bool = True,
        alpha: int = 0.001,
        verbose: bool = False,
        **kwargs: Any,
    ):
        super(WeightedAverageTransformer, self).__init__()

        if weight_type not in ["sif", "idf", "1"]:
            raise Exception("weights should be one of ['sif', 'idf', '1']")

        self.weight_type = weight_type
        self.alpha = alpha
        self.embedding_model = embedding_model
        self.embed_size = embed_size
        self.use_svd = use_svd
        self.verbose = verbose
        self.weights_ = None
        self.u_ = None
        self.w_all = 0
        self.w_emb = 0

    def get_name(self) -> str:
        """Module name.

        Returns:
            string with module name.

        """
        return self.name + "_" + self.weight_type

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            Int with module output shape.

        """
        return self.embed_size

    def reset_statistic(self):
        """Reset module statistics."""
        self.w_all = 0
        self.w_emb = 0

    def get_statistic(self):
        """Get module statistics."""
        logger.info3(f"N_words: {self.w_all}, N_emb: {self.w_emb}, coverage: {self.w_emb / self.w_all}.")

    def get_embedding_(self, sentence: Sequence[str]) -> np.ndarray:  # noqa: D102
        result = np.zeros((self.embed_size,))

        for word in sentence:
            self.w_all += 1
            if word in self.embedding_model and word in self.weights_:
                self.w_emb += 1
                result += self.embedding_model[word] * self.weights_[word]

        if len(sentence) > 0:
            result = result / len(sentence)

        return result.reshape(1, -1)

    def fit(self, sentences: Sequence[str]):  # noqa: D102
        self.reset_statistic()
        dict_vec = DictVectorizer()
        occurrences = dict_vec.fit_transform([dict(Counter(x)) for x in sentences])

        if self.weight_type == "idf":
            nd_value = np.asarray((occurrences > 0).sum(axis=0)).ravel()
            idf = np.log1p((occurrences.shape[0] + 1) / (nd_value + 1))
            self.weights_ = dict(zip(dict_vec.feature_names_, idf))

        elif self.weight_type == "sif":
            nd_value = np.asarray((occurrences > 0).sum(axis=0)).ravel()
            pw = (nd_value + 1) / (occurrences.shape[0] + 1)
            pw = self.alpha / (self.alpha + pw)
            self.weights_ = dict(zip(dict_vec.feature_names_, pw))

        else:
            self.weights_ = dict(zip(dict_vec.feature_names_, repeat(1)))

        if self.use_svd:
            if self.verbose:
                sentences = tqdm(sentences)
            sentence_embeddings = np.vstack([self.get_embedding_(x) for x in sentences])
            u, _, _ = svd(sentence_embeddings.T, full_matrices=False)
            self.u_ = u[:, 0]

        return self

    def transform(self, sentences: Sequence[str]) -> np.ndarray:  # noqa: D102
        self.reset_statistic()
        sentence_embeddings = np.vstack([self.get_embedding_(x) for x in sentences])

        if self.use_svd:
            proj = (self.u_.reshape(-1, 1) * self.u_.dot(sentence_embeddings.T).reshape(1, -1)).T
            sentence_embeddings = sentence_embeddings - proj

        return sentence_embeddings.astype(np.float32)



"""Reader utils."""

from typing import Callable
from typing import Optional
from typing import Union

import numpy as np

from sklearn.model_selection import GroupKFold
from sklearn.model_selection import KFold
from sklearn.model_selection import StratifiedKFold

from ..tasks import Task


def set_sklearn_folds(
    task: Task,
    target: np.ndarray,
    cv: Union[Callable, int] = 5,
    random_state: int = 42,
    group: Optional[np.ndarray] = None,
) -> Optional[np.ndarray]:
    """Determines the cross-validation splitting strategy.

    Args:
        task: If `'binary'` or `'multiclass'` used stratified cv.
        target: Target values.
        cv: Specifies number of folds.
        random_state: Determines random number generation.
        group: For group k-folding.

    Returns:
        Array with fold indices.

    """
    if type(cv) is int:
        if group is not None:
            split = GroupKFold(cv).split(group, group, group)
        elif task.name in ["binary", "multiclass"]:

            split = StratifiedKFold(cv, random_state=random_state, shuffle=True).split(target, target)
        else:
            split = KFold(cv, random_state=random_state, shuffle=True).split(target, target)

        folds = np.zeros(target.shape[0], dtype=np.int32)
        for n, (f0, f1) in enumerate(split):
            folds[f1] = n

        return folds

    return



"""Image feature extractors based on color histograms and CNN embeddings."""

from copy import copy
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union


try:
    import cv2
except:
    import warnings

    warnings.warn("'cv2' - package isn't installed")

import numpy as np
import torch
import torch.nn as nn


try:
    import timm

    from timm.data import resolve_data_config
    from timm.data.transforms_factory import create_transform
except:
    import warnings

    warnings.warn("'timm' - package isn't installed")

from joblib import Parallel
from joblib import delayed
from sklearn.base import TransformerMixin
from torch.utils.data import DataLoader
from tqdm import tqdm

from ..text.utils import parse_devices
from ..text.utils import seed_everything
from .utils import pil_loader


numeric = Union[int, float]


class ColorFeatures:
    """Basic class to compute color features."""

    def __init__(self, hist_size: int = 30, is_hsv: bool = True):
        """Create normalized color histogram for rgb or hsv image.

        Works with RGB and grayscale images only.

        Args:
            hist_size: Number of bins for each channel.
            is_hsv: Convert image to hsv.

        """
        self.hist_size = hist_size
        self.is_hsv = is_hsv
        self._f_names = ["h", "s", "v"] if self.is_hsv else ["r", "g", "b"]

    def compute_histogram(self, img: np.ndarray) -> List[numeric]:
        """Compute normalized color histogram for one channel.

        Args:
            img: Image with shape ``(h, w)``.

        Returns:
            List of channel histogram values.

        """
        # TODO: add value range check
        hist = cv2.calcHist([img], [0], mask=None, histSize=[self.hist_size], ranges=(0, 255))[:, 0]

        return list(hist / hist.sum())

    def get_names(self) -> List[str]:
        """Define how to get features names list.

        Returns:
            List of features names.

        """
        return [
            j
            for i in [["color_" + j + "_" + str(i) for i in np.arange(self.hist_size)] for j in self._f_names]
            for j in i
        ]

    def get_features(self, img: np.ndarray) -> List[numeric]:
        """Calculate normalized color histogram for rgb or hsv image.

        Args:
            img: Image with shape ``(h, w)``.

        Returns:
            List of histogram values.

        """
        img = copy(img)
        if self.is_hsv:
            img = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)

        res = []
        for i in range(3):
            channel = img[:, :, i]
            res.extend(self.compute_histogram(channel))

        return res


class CreateImageFeatures:
    """Class for parallel histogram computation."""

    def __init__(
        self,
        hist_size: int = 30,
        is_hsv: bool = True,
        n_jobs: int = 4,
        loader: Callable = pil_loader,
    ):
        """Create normalized color histogram for rgb or hsv image.

        Args:
            hist_size: Number of bins for each channel.
            is_hsv: Convert image to hsv.
            n_jobs: Number of threads for multiprocessing.
            loader: Callable for reading image from path.

        """
        self.fe = ColorFeatures(hist_size, is_hsv)
        self.n_jobs = n_jobs
        self.loader = loader

    def process(self, im_path_i: str) -> List[numeric]:
        """Create normalized color histogram for input image by its path.

        Args:
            im_path_i: Path to the image.

        Returns:
            List of histogram values.

        """
        im_i = np.array(self.loader(im_path_i))
        if len(im_i.shape) == 2:
            im_i = cv2.cvtColor(im_i, cv2.COLOR_GRAY2RGB)
        res_i = self.fe.get_features(im_i)
        return res_i

    def transform(self, samples: Sequence[str]) -> np.ndarray:
        """Transform input sequence with paths to histogram values.

        Args:
            samples: Sequence with images paths.

        Returns:
            Array of histograms.

        """
        res = Parallel(self.n_jobs)(delayed(self.process)(im_path_i) for im_path_i in samples)
        return np.vstack(res)


class TimmModelEmbedder(nn.Module):
    """Class to compute TimmModels embeddings."""

    def __init__(
        self,
        model_name: str = "efficientnet_b0.ra_in1k",
        weights_path: Optional[str] = None,
        device=torch.device("cuda:0"),
    ):
        """Pytorch module for image embeddings based on timm models.

        Args:
            model_name: Name of effnet model.
            weights_path: Path to saved weights.
            device: Device to use.

        """
        super(TimmModelEmbedder, self).__init__()
        self.device = device
        self.model = (
            timm.create_model(model_name, pretrained=True, num_classes=0, checkpoint_path=weights_path)
            .eval()
            .to(self.device)
        )
        self.feature_shape = self.get_shape()
        self.model_name = model_name

    @torch.no_grad()
    def get_shape(self) -> int:
        """Calculate output embedding shape.

        Returns:
            Shape of embedding.

        """
        return self.model(torch.randn(1, 3, 224, 224).to(self.device)).squeeze().shape[0]

    def forward(self, x) -> torch.Tensor:
        """Forward pass."""
        out = self.model(x)
        return out


class ImageTimmDataset:
    """Image for Timm Dataset Class."""

    def __init__(
        self,
        model: TimmModelEmbedder,
        data: Sequence[str],
        loader: Callable = pil_loader,
    ):
        """Pytorch Dataset for :class:`~lightautoml.image.TimmModelEmbedder`.

        Args:
            model: model which we train.
            data: Sequence of paths.
            loader: Callable for reading image from path.

        """
        self.X = data
        self.transforms = create_transform(**resolve_data_config(model.model.pretrained_cfg, model=model.model))
        self.loader = loader

    def __getitem__(self, idx: int) -> np.ndarray:
        path = self.X[idx]
        img = self.loader(path)
        img = self.transforms(img)
        return img

    def __len__(self):
        return len(self.X)


class DeepTimmImageEmbedder(TransformerMixin):
    """Timm Transformer for image embeddings."""

    def __init__(
        self,
        device: torch.device = torch.device("cuda:0"),
        n_jobs=4,
        random_state=42,
        model_name="efficientnet_b0.ra_in1k",
        weights_path: Optional[str] = None,
        batch_size: int = 128,
        verbose: bool = True,
    ):
        """Pytorch Dataset for :class:`~lightautoml.image.TimmModelEmbedder`.

        Args:
            device: Torch device.
            n_jobs: Number of threads for dataloader.
            random_state: Random seed.
            model_name: Name of effnet model.
            weights_path: Path to saved weights.
            batch_size: Batch size.
            verbose: Verbose data processing.

        """
        super(DeepTimmImageEmbedder, self).__init__()
        # add assert to check model
        # assert model_name in {f"efficientnet-b{i}" for i in range(8)}

        self.device, self.device_ids = parse_devices(device)
        self.random_state = random_state
        self.n_jobs = n_jobs
        self.batch_size = batch_size
        self.verbose = verbose
        seed_everything(random_state)

        self.model = TimmModelEmbedder(model_name, weights_path, self.device)

    def fit(self, data: Any = None):
        """Train model."""
        return self

    @torch.no_grad()
    def transform(self, data: Sequence[str]) -> np.ndarray:
        """Calculate image embeddings from paths.

        Args:
            data: Sequence of paths.

        Returns:
            Array of embeddings.

        """
        data = ImageTimmDataset(self.model, data)
        loader = DataLoader(data, batch_size=self.batch_size, shuffle=False, num_workers=self.n_jobs)

        result = []
        if self.verbose:
            loader = tqdm(loader)

        self.model = self.model.to(self.device)
        for batch in loader:
            embed = self.model(batch.float().to(self.device)).detach().cpu().numpy()
            result.append(embed.astype(np.float32))

        return np.vstack(result)



"""Feedforward and Attention blocks for FTTransformer (https://arxiv.org/abs/2106.11959v2) from https://github.com/lucidrains/tab-transformer-pytorch/tree/main."""

import torch
import torch.nn as nn
import torch.nn.functional as F


class GEGLU(nn.Module):
    """GEGLU activation for Attention block."""

    def forward(self, x):
        """Forward pass."""
        x, gates = x.chunk(2, dim=-1)
        return x * F.gelu(gates)


def FeedForward(dim, mult=4, dropout=0.0):
    """Feedforward for Transformer block."""
    return nn.Sequential(
        nn.LayerNorm(dim), nn.Linear(dim, dim * mult * 2), GEGLU(), nn.Dropout(dropout), nn.Linear(dim * mult, dim)
    )


class Attention(nn.Module):
    """Attention Block.

    Args:
            dim: Embeddings dimension.
            heads: Number of heads in Attention.
            dim_head: Attention head dimension.
            dropout: Post-Attention dropout.
    """

    def __init__(self, dim, heads=8, dim_head=64, dropout=0.0):
        super().__init__()
        inner_dim = dim_head * heads
        self.heads = heads
        self.scale = dim_head ** -0.5
        self.norm = nn.LayerNorm(dim)
        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)
        self.to_out = nn.Linear(inner_dim, dim, bias=False)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        """Transform the input tensor with attention.

        Args:
            x : torch.Tensor
                3-d tensor; for example, embedded numeric and/or categorical values,
                or the output of a previous attention layer.

        Returns:
            torch.Tensor

        """
        batch_size, seq_len, dim = x.shape
        h = self.heads

        x = self.norm(x)
        qkv = self.to_qkv(x)
        q, k, v = torch.split(qkv, qkv.size(-1) // 3, dim=-1)

        q = q.view(batch_size, seq_len, h, -1).permute(0, 2, 1, 3)
        k = k.view(batch_size, seq_len, h, -1).permute(0, 2, 1, 3)
        v = v.view(batch_size, seq_len, h, -1).permute(0, 2, 1, 3)

        q = q * self.scale
        sim = torch.matmul(q, k.transpose(-2, -1))
        attn = torch.softmax(sim, dim=-1)
        dropped_attn = self.dropout(attn)

        out = torch.matmul(dropped_attn, v)
        out = out.permute(0, 2, 1, 3).contiguous().view(batch_size, seq_len, -1)
        out = self.to_out(out)

        return out, attn


# transformer
class Transformer(nn.Module):
    """Transformer Block.

    Args:
            dim: Embeddings dimension.
            depth: Number of Attention Blocks.
            heads: Number of heads in Attention.
            dim_head: Attention head dimension.
            attn_dropout: Post-Attention dropout.
            ff_dropout: Feed-Forward Dropout.
            return_attn: Return attention scores or not.
    """

    def __init__(self, dim, depth, heads, dim_head, attn_dropout, ff_dropout):
        super().__init__()
        self.layers = nn.ModuleList([])

        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        Attention(dim, heads=heads, dim_head=dim_head, dropout=attn_dropout),
                        FeedForward(dim, dropout=ff_dropout),
                    ]
                )
            )

    def forward(self, x):
        """Transform the input embeddings tensor with Transformer module.

        Args:
            x : torch.Tensor
                3-d tensor; embedded numeric and/or categorical values,
                or the output of a previous Transformer layer.

        Returns:
            torch.Tensor

        """
        post_softmax_attns = []

        for attn, ff in self.layers:
            attn_out, post_softmax_attn = attn(x)
            post_softmax_attns.append(post_softmax_attn)
            x = attn_out + x
            x = ff(x) + x

        return x



"""Pytorch Datasets for text features."""

from typing import Any
from typing import Dict
from typing import Sequence
from typing import Union

import numpy as np


try:
    from transformers import AutoTokenizer
except:
    import warnings

    warnings.warn("'transformers' - package isn't installed")


class BertDataset:
    """Dataset class with transformers tokenization.

    Class for preparing transformers input.

    Args:
        sentences: List of tokenized sentences.
        max_length: Max sentence length.
        model_name: Name of transformer model.
        **kwargs: Other.

    """

    def __init__(self, sentences: Sequence[str], max_length: int, model_name: str, **kwargs: Any):
        self.sentences = sentences
        self.max_length = max_length
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast=False)

    def __getitem__(self, idx: int) -> Dict[str, np.ndarray]:
        sent = self.sentences[idx]
        _split = sent.split("[SEP]")
        sent = _split if len(_split) == 2 else (sent,)
        data = self.tokenizer.encode_plus(
            *sent, add_special_tokens=True, max_length=self.max_length, padding="max_length", truncation=True
        )
        return {i: np.array(data[i]) for i in data.keys()}

    def __len__(self) -> int:
        return len(self.sentences)


class EmbedDataset:
    """Dataset class for extracting word embeddings.

    Class for transforming list of tokens to dict of embeddings and sentence length.

    Args:
        sentences: List of tokenized sentences.
        embedding_model: word2vec, fasstext, etc.
            Should have dict interface {<word>: <embedding>}.
        max_length: Max sentence length.
        embed_size: Size of embedding.
        **kwargs: Not used.

    """

    def __init__(self, sentences: Sequence[str], embedding_model: Dict, max_length: int, embed_size: int, **kwargs):
        self.sentences = sentences
        self.embedding_model = embedding_model
        self.max_length = max_length
        self.embed_size = embed_size

    def __getitem__(self, idx: int) -> Dict[str, Union[Sequence, int]]:
        result = np.zeros((self.max_length, self.embed_size))
        length = 0
        for word in self.sentences[idx]:
            if word in self.embedding_model:
                result[length, :] = self.embedding_model[word]
                length += 1
                if length >= self.max_length:
                    break
        return {"text": result, "length": length if length > 0 else 1}

    def __len__(self) -> int:
        return len(self.sentences)



"""Text features."""

from typing import Any

from ...dataset.base import LAMLDataset
from ...text.tokenizer import BaseTokenizer
from ...text.tokenizer import SimpleEnTokenizer
from ...text.tokenizer import SimpleRuTokenizer
from ...transformers.base import ColumnsSelector
from ...transformers.base import LAMLTransformer
from ...transformers.base import SequentialTransformer
from ...transformers.base import UnionTransformer
from ...transformers.decomposition import SVDTransformer
from ...transformers.numeric import StandardScaler
from ...transformers.text import AutoNLPWrap
from ...transformers.text import ConcatTextTransformer
from ...transformers.text import TfidfTextTransformer
from ...transformers.text import TokenizerTransformer
from ..utils import get_columns_by_role
from .base import FeaturesPipeline


_model_name_by_lang = {
    "ru": "DeepPavlov/rubert-base-cased-conversational",
    "en": "bert-base-cased",
    "multi": "bert-base-multilingual-cased",
}

_tokenizer_by_lang = {
    "ru": SimpleRuTokenizer,
    "en": SimpleEnTokenizer,
    "multi": BaseTokenizer,
}


class NLPDataFeatures:
    """Class contains basic features transformations for text data."""

    _lang = {"en", "ru", "multi"}

    def __init__(self, **kwargs: Any):
        if "lang" in kwargs:
            assert kwargs["lang"] in self._lang, f"Language must be one of: {self._lang}"

        self.lang = "en" if "lang" not in kwargs else kwargs["lang"]
        self.is_tokenize_autonlp = False
        self.use_stem = False
        self.verbose = False
        self.bert_model = _model_name_by_lang[self.lang]
        self.random_state = 42
        self.device = None
        self.model_name = None
        self.embedding_model = None
        self.svd = True
        self.n_components = 100
        self.is_concat = True
        self.tfidf_params = None
        self.cache_dir = None
        self.train_fasttext = False
        self.embedding_model = None  # path to fasttext model or model with dict interface
        self.transformer_params = None  # params of random_lstm, bert_embedder, borep or wat
        self.fasttext_params = None  # init fasttext params
        self.fasttext_epochs = 2
        self.stopwords = False
        self.force = False
        self.sent_scaler = None
        self.embed_scaler = None
        # if in autonlp_params no effect
        self.multigpu = False

        for k in kwargs:
            if kwargs[k] is not None:
                self.__dict__[k] = kwargs[k]

        if not self.force and self.device == "cpu":
            self.model_name = "wat"
        else:
            if self.model_name is None:
                self.model_name = (
                    "wat"
                    if self.device == "cpu"
                    else "random_lstm"
                    if "embedding_model" in kwargs
                    else "random_lstm_bert"
                )


class TextAutoFeatures(FeaturesPipeline, NLPDataFeatures):
    """Class contains embedding features for text data."""

    def create_pipeline(self, train: LAMLDataset) -> LAMLTransformer:
        """Create pipeline for textual data.

        Args:
            train: Dataset with train features.

        Returns:
            Transformer.

        """
        transformers_list = []
        # process texts
        texts = get_columns_by_role(train, "Text")
        if len(texts) > 0:
            transforms = [ColumnsSelector(keys=texts)]
            if self.is_concat:
                transforms.append(ConcatTextTransformer())
            if self.is_tokenize_autonlp:
                transforms.append(
                    TokenizerTransformer(
                        tokenizer=_tokenizer_by_lang[self.lang](is_stemmer=self.use_stem, stopwords=self.stopwords)
                    )
                )
            transforms.append(
                AutoNLPWrap(
                    model_name=self.model_name,
                    embedding_model=self.embedding_model,
                    cache_dir=self.cache_dir,
                    bert_model=self.bert_model,
                    transformer_params=self.transformer_params,
                    random_state=self.random_state,
                    train_fasttext=self.train_fasttext,
                    device=self.device,
                    multigpu=self.multigpu,
                    sent_scaler=self.sent_scaler,
                    fasttext_params=self.fasttext_params,
                    fasttext_epochs=self.fasttext_epochs,
                    verbose=self.verbose,
                )
            )
            if self.embed_scaler == "standard":
                transforms.append(StandardScaler())

            text_processing = SequentialTransformer(transforms)
            transformers_list.append(text_processing)

        union_all = UnionTransformer(transformers_list)

        return union_all


class NLPTFiDFFeatures(FeaturesPipeline, NLPDataFeatures):
    """Class contains tfidf features for text data."""

    def create_pipeline(self, train: LAMLDataset) -> LAMLTransformer:
        """Create pipeline for textual data.

        Args:
            train: Dataset with train features.

        Returns:
            Transformer.

        """
        transformers_list = []

        # process texts
        texts = get_columns_by_role(train, "Text")
        if len(texts) > 0:
            transforms = [
                ColumnsSelector(keys=texts),
                TokenizerTransformer(
                    tokenizer=_tokenizer_by_lang[self.lang](is_stemmer=self.use_stem, stopwords=self.stopwords)
                ),
                TfidfTextTransformer(default_params=self.tfidf_params, subs=None, random_state=42),
            ]
            if self.svd:
                transforms.append(SVDTransformer(n_components=self.n_components))

            text_processing = SequentialTransformer(transforms)
            transformers_list.append(text_processing)

        union_all = UnionTransformer(transformers_list)

        return union_all


class TextBertFeatures(FeaturesPipeline, NLPDataFeatures):
    """Features pipeline for BERT."""

    def create_pipeline(self, train: LAMLDataset) -> LAMLTransformer:
        """Create pipeline for BERT.

        Args:
            train: Dataset with train data.

        Returns:
            Transformer.

        """
        transformers_list = []

        # process texts
        texts = get_columns_by_role(train, "Text")
        if len(texts) > 0:
            text_processing = SequentialTransformer(
                [
                    ColumnsSelector(keys=texts),
                    ConcatTextTransformer(),
                ]
            )
            transformers_list.append(text_processing)

        union_all = UnionTransformer(transformers_list)

        return union_all
